import datetime
from time import mktime

import six
from dateutil.relativedelta import relativedelta
from six import with_metaclass

from eventsourcing.domain.model.entity import EventSourcedEntity, EntityRepository
from eventsourcing.domain.model.events import DomainEvent, publish, QualnameABCMeta, create_domain_event_id
from eventsourcing.utils.time import timestamp_from_uuid, utc_timezone


class MessageLogged(DomainEvent):
    def __init__(self, message, entity_id, level):
        super(MessageLogged, self).__init__(entity_id=entity_id, entity_version=None, message=message, level=level)


def make_bucket_id(log_name, timestamp, bucket_size):
    d = datetime.datetime.utcfromtimestamp(timestamp)

    assert isinstance(d, datetime.datetime)
    if bucket_size.startswith('year'):
        bucket_id = '{:04}'.format(
            d.year
        )
    elif bucket_size.startswith('month'):
        bucket_id = '{:04}-{:02}'.format(
            d.year,
            d.month
        )
    elif bucket_size.startswith('day'):
        bucket_id = '{:04}-{:02}-{:02}'.format(
            d.year,
            d.month,
            d.day
        )
    elif bucket_size.startswith('hour'):
        bucket_id = '{:04}-{:02}-{:02}_{:02}'.format(
            d.year,
            d.month,
            d.day,
            d.hour
        )
    elif bucket_size.startswith('minute'):
        bucket_id = '{:04}-{:02}-{:02}_{:02}-{:02}'.format(
            d.year,
            d.month,
            d.day,
            d.hour,
            d.minute
        )
    elif bucket_size.startswith('second'):
        bucket_id = '{:04}-{:02}-{:02}_{:02}-{:02}-{:02}'.format(
            d.year,
            d.month,
            d.day,
            d.hour,
            d.minute,
            d.second
        )
    else:
        raise ValueError("Bucket size not supported: {}".format(bucket_size))
    return log_name + '_' + bucket_id


ONE_YEAR = relativedelta(years=1)
ONE_MONTH = relativedelta(months=1)
ONE_DAY = relativedelta(days=1)
ONE_HOUR = relativedelta(hours=1)
ONE_MINUTE = relativedelta(minutes=1)
ONE_SECOND = relativedelta(seconds=1)


def next_bucket_starts(timestamp, bucket_size):
    starts = bucket_starts(timestamp, bucket_size)
    duration = bucket_duration(bucket_size)
    return timestamp_from_datetime((starts + duration))


def previous_bucket_starts(timestamp, bucket_size):
    starts = bucket_starts(timestamp, bucket_size)
    duration = bucket_duration(bucket_size)
    return timestamp_from_datetime((starts - duration))


def timestamp_from_datetime(dt):
    assert dt.tzinfo, "Datetime object does not have tzinfo"
    try:
        # Only in Python 3.3+.
        return dt.timestamp()
    except AttributeError:
        return mktime(dt.timetuple()) + (dt.microsecond / 1000000.0)


def bucket_starts(timestamp, bucket_size):
    dt = datetime.datetime.utcfromtimestamp(timestamp)
    assert isinstance(dt, datetime.datetime)
    if bucket_size.startswith('year'):
        return datetime.datetime(dt.year, 1, 1, tzinfo=utc_timezone)
    elif bucket_size.startswith('month'):
        return datetime.datetime(dt.year, dt.month, 1, tzinfo=utc_timezone)
    elif bucket_size.startswith('day'):
        return datetime.datetime(dt.year, dt.month, dt.day, tzinfo=utc_timezone)
    elif bucket_size.startswith('hour'):
        return datetime.datetime(dt.year, dt.month, dt.day, dt.hour, tzinfo=utc_timezone)
    elif bucket_size.startswith('minute'):
        return datetime.datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute, tzinfo=utc_timezone)
    elif bucket_size.startswith('second'):
        return datetime.datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second, tzinfo=utc_timezone)
    else:
        raise ValueError("Bucket size not supported: {}".format(bucket_size))


BUCKET_SIZES = {
    'year': ONE_YEAR,
    'month': ONE_MONTH,
    'day': ONE_DAY,
    'hour': ONE_HOUR,
    'minute': ONE_MINUTE,
    'second': ONE_SECOND,
}


def bucket_duration(bucket_size):
    try:
        return BUCKET_SIZES[bucket_size]
    except KeyError:
        raise ValueError("Bucket size not supported: {}. Must be one of: {}"
                         "".format(bucket_size, BUCKET_SIZES.keys()))


class Log(EventSourcedEntity):

    class Started(EventSourcedEntity.Created):
        pass

    class BucketSizeChanged(EventSourcedEntity.AttributeChanged):
        pass

        @property
        def message(self):
            return self.__dict__['message']

    def __init__(self, name, bucket_size=None, **kwargs):
        super(Log, self).__init__(**kwargs)
        self._name = name
        self._bucket_size = bucket_size

    @property
    def name(self):
        return self._name

    @property
    def started_on(self):
        return self.created_on

    @property
    def bucket_size(self):
        return self._bucket_size

    def append_message(self, message, level='INFO'):
        assert isinstance(message, six.string_types)
        domain_event_id = create_domain_event_id()
        entity_bucket_id = make_bucket_id(self.name, timestamp_from_uuid(domain_event_id), self.bucket_size)
        event = MessageLogged(
            entity_id=entity_bucket_id,
            message=message,
            level=level,
        )
        publish(event)
        return event


class LogRepository(EntityRepository):
    pass


def start_new_log(name, bucket_size):
    if bucket_size not in BUCKET_SIZES:
        raise ValueError("Bucket size not supported: {}. Must be one of: {}"
                         "".format(bucket_size, BUCKET_SIZES.keys()))
    event = Log.Started(
        entity_id=name,
        name=name,
        bucket_size=bucket_size
    )
    entity = Log.mutate(event=event)
    publish(event)
    return entity


def get_logger(log):
    """
    :rtype: Logger
    """
    return Logger(log=log)


class Logger(with_metaclass(QualnameABCMeta)):

    def __init__(self, log):
        assert isinstance(log, Log), type(log)
        self.log = log

    def debug(self, message):
        return self.log.append_message(message, level='DEBUG')

    def info(self, message):
        return self.log.append_message(message, level='INFO')

    def warning(self, message):
        return self.log.append_message(message, level='WARNING')

    def error(self, message):
        return self.log.append_message(message, level='ERROR')

    def critical(self, message):
        return self.log.append_message(message, level='CRITICAL')



import unittest
from uuid import uuid1

from eventsourcing.application.subscribers.persistence import PersistenceSubscriber
from eventsourcing.domain.model.events import assert_event_handlers_empty
from eventsourcing.domain.model.example import Example, register_new_example
from eventsourcing.domain.model.snapshot import Snapshot
from eventsourcing.domain.services.eventplayer import EventPlayer
from eventsourcing.domain.services.eventstore import EventStore
from eventsourcing.domain.services.snapshotting import EventSourcedSnapshotStrategy, entity_from_snapshot, \
    take_snapshot
from eventsourcing.infrastructure.stored_event_repos.with_python_objects import PythonObjectsStoredEventRepository


class TestEventPlayer(unittest.TestCase):

    def setUp(self):
        assert_event_handlers_empty()
        self.ps = None

    def tearDown(self):
        if self.ps is not None:
            self.ps.close()
        assert_event_handlers_empty()

    def test_get_entity(self):
        # Setup an event store, using Python objects.
        event_store = EventStore(stored_event_repo=PythonObjectsStoredEventRepository())

        # Store example events.
        event1 = Example.Created(entity_id='entity1', a=1, b=2)
        event_store.append(event1)
        event2 = Example.Created(entity_id='entity2', a=2, b=4)
        event_store.append(event2)
        event3 = Example.Created(entity_id='entity3', a=3, b=6)
        event_store.append(event3)
        event4 = Example.Discarded(entity_id='entity3', entity_version=1)
        event_store.append(event4)

        # Check the event sourced entities are correct.
        # - just use a trivial mutate that always instantiates the 'Example'.
        event_player = EventPlayer(event_store=event_store, id_prefix='Example', mutate_func=Example.mutate)

        # The the reconstituted entity has correct attribute values.
        self.assertEqual('entity1', event_player.replay_events('entity1').id)
        self.assertEqual(1, event_player.replay_events('entity1').a)
        self.assertEqual(2, event_player.replay_events('entity2').a)
        self.assertEqual(None, event_player.replay_events('entity3'))

        # Check entity3 raises KeyError.
        self.assertEqual(event_player.replay_events('entity3'), None)

        # Check it works for "short" entities (should be faster, but the main thing is that it still works).
        # - just use a trivial mutate that always instantiates the 'Example'.
        event5 = Example.AttributeChanged(entity_id='entity1', entity_version=1, name='a', value=10)
        event_store.append(event5)

        event_player = EventPlayer(event_store=event_store, id_prefix='Example', mutate_func=Example.mutate)
        self.assertEqual(10, event_player.replay_events('entity1').a)

        event_player = EventPlayer(
            event_store=event_store,
            id_prefix='Example',
            mutate_func=Example.mutate,
            is_short=True,
        )
        self.assertEqual(10, event_player.replay_events('entity1').a)

    def test_snapshots(self):
        stored_event_repo = PythonObjectsStoredEventRepository()
        event_store = EventStore(stored_event_repo)
        self.ps = PersistenceSubscriber(event_store)
        event_player = EventPlayer(event_store=event_store, id_prefix='Example', mutate_func=Example.mutate)

        # Create a new entity.
        registered_example = register_new_example(a=123, b=234)

        # Take a snapshot.
        snapshot = take_snapshot(registered_example, uuid1().hex)

        # Replay from this snapshot.
        after = snapshot.at_event_id
        initial_state = entity_from_snapshot(snapshot)
        retrieved_example = event_player.replay_events(registered_example.id, initial_state=initial_state, after=after)

        # Check the attributes are correct.
        self.assertEqual(retrieved_example.a, 123)

        # Remember the time now.
        timecheck1 = uuid1().hex

        # Change attribute value.
        retrieved_example.a = 999

        # Check the initial state doesn't move.
        self.assertEqual(initial_state.a, 123)

        # Remember the time now.
        timecheck2 = uuid1().hex

        # Change attribute value.
        retrieved_example.a = 9999

        # Remember the time now.
        timecheck3 = uuid1().hex

        # Check the event sourced entities are correct.
        assert initial_state.a == 123
        retrieved_example = event_player.replay_events(registered_example.id)
        self.assertEqual(retrieved_example.a, 9999)

        # Take another snapshot.
        snapshot2 = take_snapshot(retrieved_example, uuid1().hex)

        # Check we can replay from this snapshot.
        initial_state2 = entity_from_snapshot(snapshot2)
        after2 = snapshot2.domain_event_id
        retrieved_example = event_player.replay_events(registered_example.id, initial_state=initial_state2, after=after2)
        # Check the attributes are correct.
        self.assertEqual(retrieved_example.a, 9999)

        # Check we can get historical state at timecheck1.
        retrieved_example = event_player.replay_events(registered_example.id, until=timecheck1)
        self.assertEqual(retrieved_example.a, 123)

        # Check we can get historical state at timecheck2.
        retrieved_example = event_player.replay_events(registered_example.id, until=timecheck2)
        self.assertEqual(retrieved_example.a, 999)

        # Check we can get historical state at timecheck3.
        retrieved_example = event_player.replay_events(registered_example.id, until=timecheck3)
        self.assertEqual(retrieved_example.a, 9999)

        # Similarly, check we can get historical state using a snapshot
        retrieved_example = event_player.replay_events(registered_example.id, initial_state=initial_state, after=after, until=timecheck2)
        self.assertEqual(retrieved_example.a, 999)

    def test_with_snapshots(self):
        # Check the EventPlayer's take_snapshot() method.
        stored_event_repo = PythonObjectsStoredEventRepository()
        event_store = EventStore(stored_event_repo)
        self.ps = PersistenceSubscriber(event_store)
        event_player = EventPlayer(
            event_store=event_store,
            id_prefix='Example',
            mutate_func=Example.mutate,
            snapshot_strategy=EventSourcedSnapshotStrategy(event_store=event_store)
        )

        # Check the method returns None when there are no events.
        snapshot = event_player.take_snapshot('wrong')
        self.assertIsNone(snapshot)

        # Create a new entity.
        example = register_new_example(a=123, b=234)

        # Take a snapshot with the entity.
        snapshot1 = event_player.take_snapshot(example.id)
        self.assertIsInstance(snapshot1, Snapshot)

        # Take another snapshot with the entity.
        snapshot2 = event_player.take_snapshot(example.id)
        # - should return the previous snapshot
        self.assertIsInstance(snapshot2, Snapshot)
        self.assertEqual(snapshot2.at_event_id, snapshot1.at_event_id)

        # Generate a domain event.
        example.beat_heart()

        # Take another snapshot with the entity.
        # - should use the previous snapshot and the heartbeat event
        snapshot3 = event_player.take_snapshot(example.id)
        self.assertNotEqual(snapshot3.at_event_id, snapshot1.at_event_id)



import uuid

from eventsourcing.domain.model.entity import EventSourcedEntity, mutableproperty, EntityRepository, entity_mutator, \
    singledispatch
from eventsourcing.domain.model.events import publish, DomainEvent


class Example(EventSourcedEntity):
    """
    An example event sourced domain model entity.
    """

    # Needed to get an event history longer than 10000 in Cassandra.
    __page_size__ = 1000

    # Make sure events that are applied to the entity have originated
    # from the entity at the version the instance it is current at.
    #  - this assumes _validate_originator() is called in mutators.
    __always_validate_originator_version__ = True

    class Created(EventSourcedEntity.Created):
        pass

    class AttributeChanged(EventSourcedEntity.AttributeChanged):
        pass

    class Discarded(EventSourcedEntity.Discarded):
        pass

    class Heartbeat(DomainEvent):
        pass

    def __init__(self, a, b, **kwargs):
        super(Example, self).__init__(**kwargs)
        self._a = a
        self._b = b
        self._count_heartbeats = 0

    @mutableproperty
    def a(self):
        return self._a

    @mutableproperty
    def b(self):
        return self._b

    def beat_heart(self):
        self._assert_not_discarded()
        event = self.Heartbeat(entity_id=self._id, entity_version=self._version)
        self._apply(event)
        publish(event)

    def count_heartbeats(self):
        return self._count_heartbeats

    @staticmethod
    def _mutator(event, initial):
        return example_mutator(event, initial)


@singledispatch
def example_mutator(event, initial):
    return entity_mutator(event, initial)


@example_mutator.register(Example.Heartbeat)
def heartbeat_mutator(event, self):
    self._validate_originator(event)
    assert isinstance(self, Example), self
    self._count_heartbeats += 1
    self._increment_version()
    return self


class ExampleRepository(EntityRepository):
    pass


def register_new_example(a, b):
    """
    Factory method for example entities.

    :rtype: Example
    """
    entity_id = uuid.uuid4().hex
    event = Example.Created(entity_id=entity_id, a=a, b=b)
    entity = Example.mutate(event=event)
    publish(event=event)
    return entity



from eventsourcing.exceptions import ConsistencyError, ProgrammingError
from eventsourcing.utils.time import timestamp_from_uuid

try:
    # Python 3.4+
    from functools import singledispatch
except ImportError:
    from singledispatch import singledispatch

from abc import ABCMeta, abstractmethod
from inspect import isfunction
from six import with_metaclass

from eventsourcing.domain.model.events import DomainEvent, publish, QualnameABCMeta


class EntityIDConsistencyError(ConsistencyError):
    pass


class EntityVersionConsistencyError(ConsistencyError):
    pass


class CreatedMutatorRequiresTypeNotInstance(ConsistencyError):
    pass


class EntityIsDiscarded(AssertionError):
    pass


class EventSourcedEntity(with_metaclass(QualnameABCMeta)):

    # The page size by which events are retrieved. If this
    # value is set to a positive integer, the events of
    # the entity will be retrieved in pages, using a series
    # of queries, rather than with one potentially large query.
    __page_size__ = None

    # If the entity won't have very many events, marking the entity as
    # "short" by setting __is_short__ value equal to True will mean
    # the fastest path for getting all the events is used.
    __is_short__ = False

    # This should be enabled for models that have their consistency
    # protected against concurrency errors, with e.g. optimistic
    # concurrency control. See 'always_write_entity_version' constructor
    # argument in EventSourcingApplication and StoredEventRepo classes.
    __always_validate_originator_version__ = False

    class Created(DomainEvent):
        def __init__(self, entity_version=0, **kwargs):
            super(EventSourcedEntity.Created, self).__init__(entity_version=entity_version, **kwargs)

    class AttributeChanged(DomainEvent):
        pass

    class Discarded(DomainEvent):
        pass

    def __init__(self, entity_id, entity_version, domain_event_id):
        self._id = entity_id
        self._version = entity_version
        self._is_discarded = False
        self._initial_event_id = domain_event_id

    def _increment_version(self):
        if self._version is not None:
            self._version += 1

    def _assert_not_discarded(self):
        if self._is_discarded:
            raise EntityIsDiscarded("Entity is discarded")

    @property
    def id(self):
        return self._id

    @property
    def version(self):
        return self._version

    @property
    def created_on(self):
        return timestamp_from_uuid(self._initial_event_id)

    def _validate_originator(self, event):
        # Check event's entity ID matches this entity's ID.
        if self._id != event.entity_id:
            raise EntityIDConsistencyError("Entity ID '{}' not equal to event's entity ID '{}'"
                                           "".format(self.id, event.entity_id))

        # Check event's entity version matches this entity's version.
        if self.__always_validate_originator_version__ and self._version != event.entity_version:
            raise EntityVersionConsistencyError(
                "Event version '{}' not equal to entity version '{}', event type: '{}', entity type: '{}', entity ID: '{}'"
                "".format(event.entity_version, self._version, type(event).__name__, type(self).__name__, self._id))

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def _change_attribute(self, name, value):
        self._assert_not_discarded()
        event = self.AttributeChanged(name=name, value=value, entity_id=self._id, entity_version=self._version)
        self._apply(event)
        publish(event)

    def discard(self):
        self._assert_not_discarded()
        event = self.Discarded(entity_id=self._id, entity_version=self._version)
        self._apply(event)
        publish(event)

    def _apply(self, event):
        self.mutate(event=event, entity=self)

    @classmethod
    def mutate(cls, entity=None, event=None):
        initial = entity if entity is not None else cls
        return cls._mutator(event, initial)

    @staticmethod
    def _mutator(event, initial):
        return entity_mutator(event, initial)


@singledispatch
def entity_mutator(event, _):
    raise NotImplementedError("Event type not supported: {}".format(type(event)))


@entity_mutator.register(EventSourcedEntity.Created)
def created_mutator(event, cls):
    assert isinstance(event, DomainEvent), event
    if not isinstance(cls, type):
        msg = ("Mutator for Created event requires entity type not instance: {} "
               "(event entity id: {}, event type: {})"
               "".format(type(cls), event.entity_id, type(event)))
        raise CreatedMutatorRequiresTypeNotInstance(msg)
    assert issubclass(cls, EventSourcedEntity), cls
    self = cls(**event.__dict__)
    self._increment_version()
    return self


@entity_mutator.register(EventSourcedEntity.AttributeChanged)
def attribute_changed_mutator(event, self):
    assert isinstance(self, EventSourcedEntity), self
    self._validate_originator(event)
    setattr(self, event.name, event.value)
    self._increment_version()
    return self


@entity_mutator.register(EventSourcedEntity.Discarded)
def discarded_mutator(event, self):
    assert isinstance(self, EventSourcedEntity), self
    self._validate_originator(event)
    self._is_discarded = True
    self._increment_version()
    return None


def mutableproperty(getter):
    """
    When used as a class method decorator, returns a property object
    with the method as the getter and a setter defined to call instance
    method _change_attribute(), which publishes an AttributeChanged event.
    """
    if isfunction(getter):
        def setter(self, value):
            assert isinstance(self, EventSourcedEntity), type(self)
            name = '_' + getter.__name__
            self._change_attribute(name=name, value=value)

        def new_getter(self):
            assert isinstance(self, EventSourcedEntity), type(self)
            name = '_' + getter.__name__
            return getattr(self, name)

        return property(fget=new_getter, fset=setter)
    else:
        raise ProgrammingError("Expected a function, got: {}".format(repr(getter)))


class EntityRepository(with_metaclass(ABCMeta)):

    @abstractmethod
    def __getitem__(self, entity_id):
        """
        Returns entity for given ID.
        """

    @abstractmethod
    def __contains__(self, entity_id):
        """
        Returns True or False, according to whether or not entity exists.
        """



from eventsourcing.domain.model.collection import CollectionRepository, Collection
from eventsourcing.infrastructure.event_sourced_repo import EventSourcedRepository


class CollectionRepo(EventSourcedRepository, CollectionRepository):
    """
    Event sourced repository for the Collection domain model entity.
    """
    domain_class = Collection



from uuid import uuid1

import six
from six import with_metaclass

from eventsourcing.domain.model.events import QualnameABCMeta
from eventsourcing.domain.model.log import Log, MessageLogged, make_bucket_id, next_bucket_starts, \
    previous_bucket_starts
from eventsourcing.domain.services.eventstore import AbstractEventStore
from eventsourcing.domain.services.transcoding import make_stored_entity_id
from eventsourcing.utils.time import timestamp_from_uuid


def get_log_reader(log, event_store):
    """
    :rtype: LogReader
    """
    return LogReader(log=log, event_store=event_store)


class LogReader(with_metaclass(QualnameABCMeta)):
    def __init__(self, log, event_store, page_size=50):
        assert isinstance(log, Log)
        self.log = log
        assert isinstance(event_store, AbstractEventStore)
        self.event_store = event_store
        assert isinstance(page_size, six.integer_types)
        self.page_size = page_size
        self.position = None

    def get_messages(self, after=None, until=None, limit=None, is_ascending=False, page_size=None):
        for event in self.get_events(after, until, limit, is_ascending, page_size):
            if isinstance(event, MessageLogged):
                self.position = event.domain_event_id
                yield event.message

    def get_events(self, after=None, until=None, limit=None, is_ascending=False, page_size=None):
        assert limit is None or limit > 0

        if after is None:
            after_timestamp = None
        else:
            after_timestamp = timestamp_from_uuid(after)
        if until is None:
            until_timestamp = None
        else:
            until_timestamp = timestamp_from_uuid(until)

        now_timestamp = timestamp_from_uuid(uuid1())
        started_on = self.log.started_on
        if is_ascending:
            # Start with the log start time, and continue until now.
            timestamp = started_on if after is None else max(after_timestamp, started_on)
        else:
            timestamp = now_timestamp if until is None else min(until_timestamp, now_timestamp)

        # Start counting events.
        count_events = 0

        while True:
            entity_id = make_bucket_id(self.log.name, timestamp, self.log.bucket_size)
            stored_entity_id = make_stored_entity_id('MessageLogged', entity_id)
            for message_logged_event in self.event_store.get_entity_events(
                    stored_entity_id=stored_entity_id,
                    after=after,
                    until=until,
                    limit=limit,
                    is_ascending=is_ascending,
                    page_size=page_size,
            ):
                yield message_logged_event

                if limit is not None:
                    count_events += 1
                    if count_events >= limit:
                        raise StopIteration

            # See if there's another bucket.
            if is_ascending:
                next_timestamp = next_bucket_starts(timestamp, self.log.bucket_size)
                if next_timestamp > (until_timestamp or now_timestamp):
                    raise StopIteration
                else:
                    timestamp = next_timestamp
            else:
                if timestamp < (after_timestamp or started_on):
                    raise StopIteration
                else:
                    timestamp = previous_bucket_starts(timestamp, self.log.bucket_size)



import datetime
import json
import os
import traceback
import uuid
from multiprocessing.pool import Pool
from tempfile import NamedTemporaryFile
from time import sleep
from unittest import TestCase
from uuid import uuid4, uuid1

import six

from eventsourcing.domain.services.eventstore import AbstractStoredEventRepository, SimpleStoredEventIterator
from eventsourcing.domain.services.transcoding import StoredEvent
from eventsourcing.exceptions import ConcurrencyError
from eventsourcing.infrastructure.stored_event_repos.with_cassandra import CassandraStoredEventRepository, \
    setup_cassandra_connection, get_cassandra_setup_params
from eventsourcing.infrastructure.stored_event_repos.with_python_objects import PythonObjectsStoredEventRepository
from eventsourcing.infrastructure.stored_event_repos.with_sqlalchemy import SQLAlchemyStoredEventRepository, \
    get_scoped_session_facade
from eventsourcing.infrastructure.stored_event_repos.threaded_iterator import ThreadedStoredEventIterator


class AbstractTestCase(TestCase):

    def setUp(self):
        """
        Returns None if test case class ends with 'TestCase', which means the test case isn't included in the suite.
        """
        if type(self).__name__.endswith('TestCase'):
            self.skipTest('Skipped abstract test')
        else:
            super(AbstractTestCase, self).setUp()


class AbstractStoredEventRepositoryTestCase(AbstractTestCase):

    @property
    def stored_event_repo(self):
        """
        :rtype: AbstractStoredEventRepository
        """
        raise NotImplementedError


class BasicStoredEventRepositoryTestCase(AbstractStoredEventRepositoryTestCase):

    def test_stored_event_repo(self):
        stored_entity_id = 'Entity::entity1'

        # Check the repo returns None for calls to get_most_recent_event() when there aren't any events.
        self.assertIsNone(self.stored_event_repo.get_most_recent_event(stored_entity_id))

        # Store an event for 'entity1'.
        stored_event1 = StoredEvent(event_id=uuid.uuid1().hex,
                                    stored_entity_id=stored_entity_id,
                                    event_topic='eventsourcing.domain.model.example#Example.Created',
                                    event_attrs='{"a":1,"b":2,"stored_entity_id":"entity1","timestamp":3}')
        self.stored_event_repo.append(stored_event1)

        # Store another event for 'entity1'.
        stored_event2 = StoredEvent(event_id=uuid.uuid1().hex,
                                    stored_entity_id=stored_entity_id,
                                    event_topic='eventsourcing.domain.model.example#Example.Created',
                                    event_attrs='{"a":1,"b":2,"stored_entity_id":"entity1","timestamp":4}')
        self.stored_event_repo.append(stored_event2)

        # Get all events for 'entity1'.
        retrieved_events = self.stored_event_repo.get_entity_events(stored_entity_id)
        retrieved_events = list(retrieved_events)  # Make sequence from the iterator.

        num_fixture_events = 2

        self.assertEqual(num_fixture_events, len(retrieved_events))
        # - check the first event
        self.assertIsInstance(retrieved_events[0], StoredEvent)
        self.assertEqual(stored_event1.event_topic, retrieved_events[0].event_topic)
        self.assertEqual(stored_event1.event_attrs, retrieved_events[0].event_attrs)
        # - check the second event
        self.assertIsInstance(retrieved_events[1], StoredEvent)
        self.assertEqual(stored_event2.event_topic, retrieved_events[1].event_topic)
        self.assertEqual(stored_event2.event_attrs, retrieved_events[1].event_attrs)

        # Get with different combinations of query and result ascending or descending.
        retrieved_events = self.stored_event_repo.get_entity_events(stored_entity_id, query_ascending=True, results_ascending=True)
        retrieved_events = list(retrieved_events)
        self.assertEqual(stored_event1.event_attrs, retrieved_events[0].event_attrs)
        self.assertEqual(stored_event2.event_attrs, retrieved_events[1].event_attrs)

        retrieved_events = self.stored_event_repo.get_entity_events(stored_entity_id, query_ascending=True, results_ascending=False)
        retrieved_events = list(retrieved_events)
        self.assertEqual(stored_event1.event_attrs, retrieved_events[1].event_attrs)
        self.assertEqual(stored_event2.event_attrs, retrieved_events[0].event_attrs)

        retrieved_events = self.stored_event_repo.get_entity_events(stored_entity_id, query_ascending=False, results_ascending=True)
        retrieved_events = list(retrieved_events)
        self.assertEqual(stored_event1.event_attrs, retrieved_events[0].event_attrs)
        self.assertEqual(stored_event2.event_attrs, retrieved_events[1].event_attrs)

        retrieved_events = self.stored_event_repo.get_entity_events(stored_entity_id, query_ascending=False, results_ascending=False)
        retrieved_events = list(retrieved_events)
        self.assertEqual(stored_event1.event_attrs, retrieved_events[1].event_attrs)
        self.assertEqual(stored_event2.event_topic, retrieved_events[0].event_topic)

        # Get with limit (depends on query order).
        retrieved_events = self.stored_event_repo.get_entity_events(stored_entity_id, limit=1, query_ascending=True)
        retrieved_events = list(retrieved_events)  # Make sequence from the iterator.
        # - check the first retrieved event is the first event that was stored
        self.assertIsInstance(retrieved_events[0], StoredEvent)
        self.assertEqual(stored_event1.event_topic, retrieved_events[0].event_topic)
        self.assertEqual(stored_event1.event_attrs, retrieved_events[0].event_attrs)

        retrieved_events = self.stored_event_repo.get_entity_events(stored_entity_id, limit=1, query_ascending=False)
        retrieved_events = list(retrieved_events)  # Make sequence from the iterator.
        # - check the first retrieved event is the last event that was stored
        self.assertIsInstance(retrieved_events[0], StoredEvent)
        self.assertEqual(stored_event2.event_topic, retrieved_events[0].event_topic)
        self.assertEqual(stored_event2.event_attrs, retrieved_events[0].event_attrs)

        # Get the most recent event for 'entity1'.
        most_recent_event = self.stored_event_repo.get_most_recent_event(stored_entity_id)
        self.assertIsInstance(most_recent_event, StoredEvent)
        self.assertEqual(most_recent_event.event_id, stored_event2.event_id)

        # Get all events for 'entity1' since the first event's timestamp.
        retrieved_events = self.stored_event_repo.get_entity_events(stored_entity_id, after=stored_event1.event_id)
        retrieved_events = list(retrieved_events)  # Make sequence from the iterator.
        self.assertEqual(1, len(list(retrieved_events)))
        # - check the last event is first
        self.assertIsInstance(retrieved_events[0], StoredEvent)
        self.assertEqual(stored_event2.event_topic, retrieved_events[0].event_topic)
        self.assertEqual(stored_event2.event_attrs, retrieved_events[0].event_attrs)

        # Store lots of events, and check the latest events can be retrieved quickly.
        stored_events = []
        num_extra_events = 127
        for page_count in six.moves.range(num_extra_events):
            stored_event_i = StoredEvent(event_id=uuid.uuid1().hex,
                                         stored_entity_id=stored_entity_id,
                                         event_topic='eventsourcing.domain.model.example#Example.Created',
                                         event_attrs='{"a":1,"b":2,"stored_entity_id":"entity1","timestamp":%s}' % (page_count+10))
            stored_events.append(stored_event_i)
            self.stored_event_repo.append(stored_event_i)

        last_snapshot_event = stored_events[-20]

        # start_time = datetime.datetime.now()
        retrieved_events = self.stored_event_repo.get_entity_events(stored_entity_id, after=last_snapshot_event.event_id)
        retrieved_events = list(retrieved_events)
        # page_duration = (datetime.datetime.now() - start_time).total_seconds()
        # self.assertLess(page_duration, 0.05)
        # print("Duration: {}".format(page_duration))

        self.assertEqual(len(retrieved_events), 19)
        self.assertIsInstance(retrieved_events[-1], StoredEvent)
        self.assertEqual(stored_events[-19].event_topic, retrieved_events[0].event_topic)
        self.assertEqual(stored_events[-19].event_attrs, retrieved_events[0].event_attrs)
        self.assertEqual(stored_events[-1].event_topic, retrieved_events[-1].event_topic)
        self.assertEqual(stored_events[-1].event_attrs, retrieved_events[-1].event_attrs)

        # Check the first events can be retrieved easily.
        start_time = datetime.datetime.now()
        retrieved_events = self.stored_event_repo.get_entity_events(stored_entity_id, limit=20, query_ascending=True)
        retrieved_events = list(retrieved_events)
        page_duration = (datetime.datetime.now() - start_time).total_seconds()
        # print("Duration: {}".format(page_duration))
        # self.assertLess(page_duration, 0.05)
        self.assertEqual(len(retrieved_events), 20)
        self.assertIsInstance(retrieved_events[0], StoredEvent)
        self.assertEqual(stored_event1.event_topic, retrieved_events[0].event_topic)
        self.assertEqual(stored_event1.event_attrs, retrieved_events[0].event_attrs)

        # Check the next page of events can be retrieved easily.
        start_time = datetime.datetime.now()
        retrieved_events = self.stored_event_repo.get_entity_events(stored_entity_id, after=retrieved_events[-1].event_id,
                                                               limit=20, query_ascending=True)
        retrieved_events = list(retrieved_events)
        page_duration = (datetime.datetime.now() - start_time).total_seconds()
        # self.assertLess(page_duration, 0.05)
        # print("Duration: {}".format(page_duration))

        self.assertEqual(len(retrieved_events), 20)
        self.assertIsInstance(retrieved_events[0], StoredEvent)
        self.assertEqual(stored_events[18].event_topic, retrieved_events[0].event_topic)
        self.assertEqual(stored_events[18].event_attrs, retrieved_events[0].event_attrs)


class ConcurrentStoredEventRepositoryTestCase(AbstractStoredEventRepositoryTestCase):

    def setUp(self):
        super(ConcurrentStoredEventRepositoryTestCase, self).setUp()
        self.app = None
        self.temp_file = NamedTemporaryFile('a')

    def tearDown(self):
        super(ConcurrentStoredEventRepositoryTestCase, self).tearDown()
        if self.app is not None:
            self.app.close()

    def test_optimistic_concurrency_control(self):
        """Appends lots of events, but with a pool of workers
        all trying to add the same sequence of events.
        """
        # Start a pool.
        pool_size = 3
        print("Pool size: {}".format(pool_size))
        pool = Pool(
            initializer=pool_initializer,
            processes=pool_size,
            initargs=(type(self.stored_event_repo), self.temp_file.name),
        )

        # Append duplicate events to the repo, or at least try...
        number_of_events = 40
        self.assertGreater(number_of_events, pool_size)
        stored_entity_id = uuid4().hex
        sequence_of_args = [(number_of_events, stored_entity_id)] * pool_size
        results = pool.map(append_lots_of_events_to_repo, sequence_of_args)
        total_successes = []
        total_failures = []
        for result in results:
            if isinstance(result, Exception):
                print(result.args[0][1])
                raise result
            else:
                successes, failures = result
                assert isinstance(successes, list), result
                assert isinstance(failures, list), result
                total_successes.extend(successes)
                total_failures.extend(failures)

        # Close the pool.
        pool.close()

        # Check each event version was written exactly once.
        self.assertEqual(sorted([i[0] for i in total_successes]), list(range(number_of_events)))

        # Check each child got to write at least one event.
        self.assertEqual(len(set([i[1] for i in total_successes])), pool_size)

        # Check each event version at least once wasn't written due to a concurrency error.
        self.assertEqual(sorted(set(sorted([i[0] for i in total_failures]))), list(range(number_of_events)))

        # Check at least one event version wasn't written due to a concurrency error.
        self.assertTrue(set([i[0] for i in total_failures]))

        # Check each child failed to write at least one event.
        self.assertEqual(len(set([i[1] for i in total_failures])), pool_size)

        # Check there's actually a contiguous version sequence through the sequence of stored events.
        events = self.stored_event_repo.get_entity_events(stored_entity_id)
        self.assertEqual(len(events), number_of_events)
        version_counter = 0
        for event in events:
            assert isinstance(event, StoredEvent)
            attr_values = json.loads(event.event_attrs)
            self.assertEqual(attr_values['entity_version'], version_counter)
            version_counter += 1

        # Join the pool.
        pool.join()

    @staticmethod
    def append_lots_of_events_to_repo(args):

        num_events_to_create, stored_entity_id = args

        success_count = 0
        assert isinstance(worker_repo, AbstractStoredEventRepository)
        assert isinstance(num_events_to_create, six.integer_types)

        successes = []
        failures = []

        try:

            while True:
                # Imitate an entity getting refreshed, by getting the version of the last event.
                events = worker_repo.get_entity_events(stored_entity_id, limit=1, query_ascending=False)
                if len(events):
                    current_version = json.loads(events[0].event_attrs)['entity_version']
                    new_version = current_version + 1
                else:
                    current_version = None
                    new_version = 0

                # Stop before the version number gets too high.
                if new_version >= num_events_to_create:
                    break

                pid = os.getpid()
                try:

                    # Append an event.
                    stored_event = StoredEvent(
                        event_id=uuid1().hex,
                        stored_entity_id=stored_entity_id,
                        event_topic='topic',
                        event_attrs=json.dumps({'a': 1, 'b': 2, 'entity_version': new_version}),
                    )
                    started = datetime.datetime.now()
                    worker_repo.append(
                        new_stored_event=stored_event,
                        new_version_number=new_version,
                        max_retries=10,
                        artificial_failure_rate=0.25,
                    )
                except ConcurrencyError:
                    # print("PID {} got concurrent exception writing event at version {} at {}".format(
                    #     pid, new_version, started, datetime.datetime.now() - started))
                    failures.append((new_version, pid))
                    sleep(0.01)
                else:
                    print("PID {} success writing event at version {} at {} in {}".format(
                        pid, new_version, started, datetime.datetime.now() - started))
                    success_count += 1
                    successes.append((new_version, pid))
                    # Delay a successful writer, to give other processes a chance to write the next event.
                    sleep(0.03)

        # Return to parent process the successes and failure, or an exception.
        except Exception as e:
            msg = traceback.format_exc()
            print(" - failed to append event: {}".format(msg))
            return Exception((e, msg))
        else:
            return (successes, failures)


def pool_initializer(stored_repo_class, temp_file_name):
    global worker_repo

    repo = create_repo(stored_repo_class, temp_file_name)
    worker_repo = repo


def append_lots_of_events_to_repo(args):
    return ConcurrentStoredEventRepositoryTestCase.append_lots_of_events_to_repo(args)


worker_repo = None


def create_repo(stored_repo_class, temp_file_name):
    if stored_repo_class == CassandraStoredEventRepository:
        setup_cassandra_connection(*get_cassandra_setup_params())
        repo = CassandraStoredEventRepository(
            always_check_expected_version=True,
            always_write_entity_version=True,
        )
    elif stored_repo_class == SQLAlchemyStoredEventRepository:
        uri = 'sqlite:///' + temp_file_name
        scoped_session_facade = get_scoped_session_facade(uri)
        repo = SQLAlchemyStoredEventRepository(scoped_session_facade,
            always_check_expected_version=True,
            always_write_entity_version=True,
        )
    elif stored_repo_class == PythonObjectsStoredEventRepository:
        repo = PythonObjectsStoredEventRepository(
            always_check_expected_version=True,
            always_write_entity_version=True,
        )
    else:
        raise Exception("Stored repo class not yet supported in test: {}".format(stored_repo_class))
    return repo


class IteratorTestCase(AbstractStoredEventRepositoryTestCase):

    @property
    def stored_entity_id(self):
        return 'Entity::1'

    @property
    def num_events(self):
        return 12

    @property
    def iterator_cls(self):
        """
        Returns iterator class.
        """
        raise NotImplementedError

    def create_iterator(self, is_ascending, page_size):
        return self.iterator_cls(
            repo=self.stored_event_repo,
            stored_entity_id=self.stored_entity_id,
            page_size=page_size,
            is_ascending=is_ascending,
        )

    def setup_stored_events(self):
        self.stored_events = []
        self.number_of_stored_events = 12
        for page_number in six.moves.range(self.number_of_stored_events):
            stored_event = StoredEvent(
                event_id=uuid.uuid1().hex,
                stored_entity_id=self.stored_entity_id,
                event_topic='eventsourcing.domain.model.example#Example.Created',
                event_attrs='{"a":%s,"b":2,"stored_entity_id":"%s","timestamp":%s}' % (
                    page_number, self.stored_entity_id, uuid1().hex
                )
            )
            self.stored_events.append(stored_event)
            self.stored_event_repo.append(stored_event)

    def test(self):
        self.setup_stored_events()

        assert isinstance(self.stored_event_repo, AbstractStoredEventRepository)
        stored_events = self.stored_event_repo.get_entity_events(stored_entity_id=self.stored_entity_id)
        stored_events = list(stored_events)
        self.assertEqual(len(stored_events), self.num_events)

        # Check can get all events.
        page_size = 5

        # Iterate ascending.
        is_ascending = True
        expect_at_start = self.stored_events[0].event_attrs
        expect_at_end = self.stored_events[-1].event_attrs
        self.assert_iterator_yields_events(is_ascending, expect_at_start, expect_at_end, page_size)

        # Iterate descending.
        is_ascending = False
        expect_at_start = self.stored_events[-1].event_attrs
        expect_at_end = self.stored_events[0].event_attrs
        self.assert_iterator_yields_events(is_ascending, expect_at_start, expect_at_end, page_size)

    def assert_iterator_yields_events(self, is_ascending, expect_at_start, expect_at_end, page_size):
        iterator = self.create_iterator(is_ascending, page_size)
        retrieved_events = list(iterator)
        self.assertEqual(len(retrieved_events), len(self.stored_events))
        self.assertGreater(len(retrieved_events), page_size)
        self.assertEqual(iterator.page_counter, 3)
        self.assertEqual(iterator.all_event_counter, self.num_events)
        self.assertEqual(expect_at_start, retrieved_events[0].event_attrs)
        self.assertEqual(expect_at_end, retrieved_events[-1].event_attrs)


class SimpleStoredEventIteratorTestCase(IteratorTestCase):

    @property
    def iterator_cls(self):
        return SimpleStoredEventIterator


class ThreadedStoredEventIteratorTestCase(IteratorTestCase):

    @property
    def iterator_cls(self):
        return ThreadedStoredEventIterator


from abc import abstractmethod, ABCMeta

from six import with_metaclass

from eventsourcing.application.subscribers.persistence import PersistenceSubscriber
from eventsourcing.domain.services.eventstore import EventStore
from eventsourcing.domain.services.transcoding import Transcoder


class EventSourcingApplication(with_metaclass(ABCMeta)):
    persist_events = True

    def __init__(self, json_encoder_cls=None, json_decoder_cls=None, cipher=None, always_encrypt_stored_events=False,
                 enable_occ=False, always_write_entity_version=False):
        """
        Initialises event sourcing application attributes. Constructs a stored event repo using a
        concrete method that must be provided by a subclass, an event store using the stored
        event repository, and optionally a persistence subscriber that uses the event store.

        As well as providing a concrete method to construct a stored event repository, subclasses
        may use the event store to construct event sourced repositories, from which event sourced
        entities can be retrieved. Subclasses may also construct other subscribers which excute
        commands against the entities in response to the publication of domain events.

        To enable symmetric encryption of stored events, pass in a 'cipher' and
        provide a True value for 'always_encrypt_stored_events'.

        To enable optimistic concurrency control (new in 1.1.0), pass a True value for argument
        'always_check_expected_version'. This feature depends on your having a table
        'entity_versions' is your schema, so if you are upgrading from v1.0.x of this package,
        then please check your schema and write a migration script for your application before
        deploying version 1.1.x into production.


        :param json_encoder_cls:  JSON encoder class.

        :param json_decoder_cls:  JSON decoder class.

        :param cipher:  Encryption cypher for encryption of event attributes when stored.

        :param always_encrypt_stored_events:  Apply encryption to all stored events.

        :param enable_occ:  Enables optimistic concurrency control, so that an exception
                            is raised when writing versions that would be out of order. This
                            also enabled writing of entity versions, which is required for
                            optimistic concurrency control.

        :param always_write_entity_version: Enables writing of entity versions, which is
                                            required for optimistic concurrency control.
                                            In itself, this option doesn't enable optimistic
                                            concurrency control, but may help to achieve a
                                            zero-downtime migration for an application
                                            developed with a previous version of this package.

        Suggested steps for migration from one-table schema (v1.0.x) to the new two table schema (v1.1.x):

        - upgrade the application to use the new version of this package
          but don't enable any of the optiions

        - after testing, deploy with the new version of this package, and then migrate
          the database to add the new 'entity_versions' table

        - change the application to enable writing entity versions

        - after testing, deploy the new version

        - write a script to add to the entity versions table: at one event for each
          stored entity which corresponds to the very last event for that entity; or
          for each stored event as it would be if this feature had existed all along

        - change the application to enable optimistic concurrency control - you
          will also need to change commands that are contentious in your application
          to retry the command, to get a fresh version of the entity repeat the
          operation

        - after testing, deploy the new version - your versioned event streams are
          now protected by optimistic concurrency control, so that they will not
          become inconsistent, please note that events that are not versioned do not
          have optimistic concurrency controls applied to them, since there is nothing
          to control
        """
        self.stored_event_repo = self.create_stored_event_repo(
            always_check_expected_version=enable_occ,
            always_write_entity_version=enable_occ or always_write_entity_version,
        )
        self.event_store = self.create_event_store(
            json_encoder_cls=json_encoder_cls, json_decoder_cls=json_decoder_cls,
            cipher=cipher, always_encrypt=always_encrypt_stored_events,

        )
        self.persistence_subscriber = self.create_persistence_subscriber()

    @abstractmethod
    def create_stored_event_repo(self, **kwargs):
        """Returns an instance of a subclass of AbstractStoredEventRepository.

        :rtype: AbstractStoredEventRepository
        """

    def create_event_store(self, json_encoder_cls=None, json_decoder_cls=None, cipher=None, always_encrypt=False):
        transcoder = self.create_transcoder(always_encrypt, cipher, json_decoder_cls, json_encoder_cls)
        return EventStore(
            stored_event_repo=self.stored_event_repo,
            transcoder=transcoder,
        )

    def create_transcoder(self, always_encrypt, cipher, json_decoder_cls, json_encoder_cls):
        return Transcoder(
            json_encoder_cls=json_encoder_cls,
            json_decoder_cls=json_decoder_cls,
            cipher=cipher,
            always_encrypt=always_encrypt,
        )

    def create_persistence_subscriber(self):
        if self.persist_events and self.event_store:
            return PersistenceSubscriber(
                event_store=self.event_store
            )

    def close(self):
        if self.persistence_subscriber is not None:
            self.persistence_subscriber.close()
            self.persistence_subscriber = None
        self.event_store = None
        self.stored_event_repo = None

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.close()



from tempfile import NamedTemporaryFile
from unittest import TestCase

from eventsourcing.infrastructure.stored_event_repos.with_sqlalchemy import get_scoped_session_facade, \
    SQLAlchemyStoredEventRepository


class SQLAlchemyTestCase(TestCase):

    @property
    def stored_event_repo(self):
        try:
            return self._stored_event_repo
        except AttributeError:
            self.temp_file = NamedTemporaryFile('a')
            uri = 'sqlite:///' + self.temp_file.name
            scoped_session_facade = get_scoped_session_facade(uri)
            stored_event_repo = SQLAlchemyStoredEventRepository(scoped_session_facade)
            self._stored_event_repo = stored_event_repo
            return self._stored_event_repo

    def tearDown(self):
        # Unlink temporary file.
        if self.temp_file:
            self.temp_file.close()
        super(SQLAlchemyTestCase, self).tearDown()


from abc import ABCMeta, abstractmethod

import six

from eventsourcing.domain.model.events import topic_from_domain_class, publish
from eventsourcing.domain.model.snapshot import Snapshot, AbstractSnapshop
from eventsourcing.domain.services.eventstore import AbstractEventStore
from eventsourcing.domain.services.transcoding import make_stored_entity_id, id_prefix_from_event_class, \
    deserialize_domain_entity, id_prefix_from_entity


class AbstractSnapshotStrategy(six.with_metaclass(ABCMeta)):

    @abstractmethod
    def get_snapshot(self, stored_entity_id, until=None):
        """Returns pre-existing snapshot for stored entity ID from given
        event store, optionally until a particular domain event ID.
        """

    @abstractmethod
    def take_snapshot(self, entity, at_event_id):
        """Creates snapshot from given entity, with given domain event ID.
        """


class EventSourcedSnapshotStrategy(AbstractSnapshotStrategy):
    """Snapshot strategy that uses an event sourced snapshot.
    """
    def __init__(self, event_store):
        self.event_store = event_store

    def get_snapshot(self, stored_entity_id, until=None):
        return get_snapshot(stored_entity_id, self.event_store, until=until)

    def take_snapshot(self, entity, at_event_id):
        return take_snapshot(entity=entity, at_event_id=at_event_id)


def get_snapshot(stored_entity_id, event_store, until=None):
    """
    Get the last snapshot for entity.

    :rtype: Snapshot
    """
    assert isinstance(event_store, AbstractEventStore)
    snapshot_entity_id = make_stored_entity_id(id_prefix_from_event_class(Snapshot), stored_entity_id)
    return event_store.get_most_recent_event(snapshot_entity_id, until=until)


def entity_from_snapshot(snapshot):
    """Deserialises a domain entity from a snapshot object.
    """
    assert isinstance(snapshot, AbstractSnapshop)
    return deserialize_domain_entity(snapshot.topic, snapshot.attrs)


def take_snapshot(entity, at_event_id):
    # Make the 'stored entity ID' for the entity, it is used as the Snapshot 'entity ID'.
    id_prefix = id_prefix_from_entity(entity)
    stored_entity_id = make_stored_entity_id(id_prefix, entity.id)

    # Create the snapshot event.
    snapshot = Snapshot(
        entity_id=stored_entity_id,
        domain_event_id=at_event_id,
        topic=topic_from_domain_class(entity.__class__),
        attrs=entity.__dict__.copy(),
    )
    publish(snapshot)

    # Return the event.
    return snapshot



import datetime
from unittest.case import TestCase

from six import with_metaclass

from eventsourcing.domain.model.events import DomainEvent, topic_from_domain_class, QualnameABCMeta
from eventsourcing.domain.model.example import Example
from eventsourcing.domain.services.transcoding import serialize_domain_event, deserialize_domain_event, \
    resolve_domain_topic, StoredEvent, ObjectJSONDecoder, ObjectJSONEncoder
from eventsourcing.exceptions import TopicResolutionError
from eventsourcing.utils.time import utc_timezone


class TestStoredEvent(TestCase):

    def test_serialize_domain_event(self):
        datetime_now = datetime.datetime(2015, 9, 8, 16, 20, 50, 577429)
        datetime_now_tzaware = datetime.datetime(2015, 9, 8, 16, 20, 50, 577429, tzinfo=utc_timezone)
        date_now = datetime.date(2015, 9, 8)
        event1 = DomainEvent(a=1, b=2, c=datetime_now, d=datetime_now_tzaware, e=date_now, entity_version=0, entity_id='entity1', domain_event_id=3)
        stored_event = serialize_domain_event(event1, json_encoder_cls=ObjectJSONEncoder)
        self.assertEqual('DomainEvent::entity1', stored_event.stored_entity_id)
        self.assertEqual('eventsourcing.domain.model.events#DomainEvent', stored_event.event_topic)
        self.assertEqual('{"a":1,"b":2,"c":{"ISO8601_datetime":"2015-09-08T16:20:50.577429"},"d":{"ISO8601_datetime":"2015-09-08T16:20:50.577429+0000"},"domain_event_id":3,"e":{"ISO8601_date":"2015-09-08"},"entity_id":"entity1","entity_version":0}',
                         stored_event.event_attrs)

    def test_serialize_domain_event_with_numpy_array(self):
        try:
            import numpy
        except ImportError:
            numpy = None

        if numpy is not None:
            event1 = DomainEvent(a=numpy.array([10.123456]), entity_version=0, entity_id='entity1', domain_event_id=3)

            stored_event = serialize_domain_event(event1, json_encoder_cls=ObjectJSONEncoder)
            self.assertEqual('eventsourcing.domain.model.events#DomainEvent', stored_event.event_topic)
            self.assertEqual('{"a":{"__ndarray__":"\\"\\\\u0093NUMPY\\\\u0001\\\\u0000F\\\\u0000{\'descr\': \'<f8\', \'fortran_order\': False, \'shape\': (1,), }            \\\\nm\\\\u00fd\\\\u00f4\\\\u009f5?$@\\""},"domain_event_id":3,"entity_id":"entity1","entity_version":0}',
                             stored_event.event_attrs)
        else:
            self.skipTest("Skipped test because numpy is not installed")

    def test_recreate_domain_event(self):
        stored_event = StoredEvent(event_id='1',
                                   stored_entity_id='entity1',
                                   event_topic='eventsourcing.domain.model.events#DomainEvent',
                                   event_attrs='{"a":1,"b":2,"c":{"ISO8601_datetime":"2015-09-08T16:20:50.577429"},"d":{"ISO8601_datetime":"2015-09-08T16:20:50.577429+0000"},"domain_event_id":3,"e":{"ISO8601_date":"2015-09-08"},"entity_id":"entity1","entity_version":0}')
        domain_event = deserialize_domain_event(stored_event, json_decoder_cls=ObjectJSONDecoder)
        self.assertIsInstance(domain_event, DomainEvent)
        self.assertEqual('entity1', domain_event.entity_id)
        self.assertEqual(1, domain_event.a)
        self.assertEqual(2, domain_event.b)
        datetime_now_timezoneaware = datetime.datetime(2015, 9, 8, 16, 20, 50, 577429, tzinfo=utc_timezone)
        # self.assertEqual(datetime_now, domain_event.c)
        self.assertEqual(datetime_now_timezoneaware, domain_event.d)
        date_now = datetime.date(2015, 9, 8)
        self.assertEqual(date_now, domain_event.e)
        self.assertEqual(3, domain_event.domain_event_id)

        # Check the TypeError is raised.
        stored_event = StoredEvent(event_id='1',
                                   stored_entity_id='entity1',
                                   event_topic=topic_from_domain_class(NotADomainEvent),
                                   event_attrs='{"a":1,"b":2,"stored_entity_id":"entity1","timestamp":3}')
        self.assertRaises(ValueError, deserialize_domain_event, stored_event, json_decoder_cls=ObjectJSONDecoder)

    def test_resolve_event_topic(self):
        example_topic = 'eventsourcing.domain.model.example#Example.Created'
        actual = resolve_domain_topic(example_topic)
        self.assertEqual(Example.Created, actual)
        example_topic = 'xxxxxxxxxxxxx#Example.Event'
        self.assertRaises(TopicResolutionError, resolve_domain_topic, example_topic)
        example_topic = 'eventsourcing.domain.model.example#Xxxxxxxx.Xxxxxxxx'
        self.assertRaises(TopicResolutionError, resolve_domain_topic, example_topic)


class NotADomainEvent(with_metaclass(QualnameABCMeta)):
    pass



from eventsourcing.application.example.with_sqlalchemy import ExampleApplicationWithSQLAlchemy
from eventsourcing.tests.unit_test_cases_example_application import ExampleApplicationTestCase


class TestApplicationWithSQLAlchemy(ExampleApplicationTestCase):

    def create_app(self):
        return ExampleApplicationWithSQLAlchemy(db_uri='sqlite:///:memory:')



from abc import abstractproperty

from eventsourcing.domain.model.entity import EntityRepository, EventSourcedEntity
from eventsourcing.domain.services.eventplayer import EventPlayer
from eventsourcing.domain.services.eventstore import AbstractEventStore
from eventsourcing.domain.services.snapshotting import entity_from_snapshot
from eventsourcing.domain.services.transcoding import id_prefix_from_entity_class
from eventsourcing.exceptions import RepositoryKeyError


class EventSourcedRepository(EntityRepository):

    # If the entity won't have very many events, marking the entity as
    # "short" by setting __is_short__ value equal to True will mean
    # the fastest path for getting all the events is used. If you set
    # a value for page size (see below), this option will have no effect.
    __is_short__ = False

    # The page size by which events are retrieved. If this
    # value is set to a positive integer, the events of
    # the entity will be retrieved in pages, using a series
    # of queries, rather than with one potentially large query.
    __page_size__ = None

    def __init__(self, event_store, use_cache=False, snapshot_strategy=None):
        self._cache = {}
        self._use_cache = use_cache
        self._snapshot_strategy = snapshot_strategy

        # Check we got an event store.
        assert isinstance(event_store, AbstractEventStore)
        self.event_store = event_store

        # Check domain class is a type of event sourced entity.
        assert issubclass(self.domain_class, EventSourcedEntity)

        # Instantiate an event player for this repo, with
        # repo-specific mutate function, page size, etc.
        self.event_player = EventPlayer(
            event_store=self.event_store,
            id_prefix=id_prefix_from_entity_class(self.domain_class),
            mutate_func=self.domain_class.mutate,
            page_size=self.__page_size__ or self.domain_class.__page_size__,
            is_short=self.__is_short__ or self.domain_class.__is_short__,
            snapshot_strategy=self._snapshot_strategy,
        )

    def __contains__(self, entity_id):
        """
        Returns a boolean value according to whether entity with given ID exists.
        """
        return self.get_entity(entity_id) is not None

    def __getitem__(self, entity_id):
        """
        Returns entity with given ID.
        """
        # Get entity from the cache.
        if self._use_cache:
            try:
                return self._cache[entity_id]
            except KeyError:
                pass

        # Reconstitute the entity.
        entity = self.get_entity(entity_id)

        # Never created or already discarded?
        if entity is None:
            raise RepositoryKeyError(entity_id)

        # Put entity in the cache.
        if self._use_cache:
            self.add_cache(entity_id, entity)

        # Return entity.
        return entity

    def add_cache(self, entity_id, entity):
        self._cache[entity_id] = entity

    @abstractproperty
    def domain_class(self):
        """
        Returns the type of entity held by this repository.
        """
        return EventSourcedEntity

    def get_entity(self, entity_id, until=None):
        """
        Returns entity with given ID, optionally as it was until the given domain event ID.
        """

        # Get a snapshot (None if none exist).
        snapshot = self.event_player.get_snapshot(entity_id, until)

        # Decide the initial state, and after when we need to get the events.
        if snapshot is None:
            after = None
            initial_state = None
        else:
            after = snapshot.at_event_id
            initial_state = entity_from_snapshot(snapshot)

        # Replay domain events.
        return self.event_player.replay_events(entity_id, after=after, until=until, initial_state=initial_state)

    def fastforward(self, stale_entity, until=None):
        """
        Mutates an instance of an entity, according to the events that have occurred since its version.
        """
        return self.event_player.fastforward(stale_entity, until=until)



import datetime
from time import sleep
from uuid import uuid1

from eventsourcing.application.subscribers.persistence import PersistenceSubscriber
from eventsourcing.domain.model.events import assert_event_handlers_empty
from eventsourcing.domain.model.log import get_logger, Logger, start_new_log, Log
from eventsourcing.domain.services.eventstore import EventStore
from eventsourcing.infrastructure.event_sourced_repos.log_repo import LogRepo
from eventsourcing.infrastructure.log_reader import get_log_reader
from eventsourcing.tests.unit_test_cases import AbstractTestCase
from eventsourcing.tests.unit_test_cases_cassandra import CassandraStoredEventRepoTestCase
from eventsourcing.tests.unit_test_cases_python_objects import PythonObjectsTestCase
from eventsourcing.tests.unit_test_cases_sqlalchemy import SQLAlchemyTestCase


class LogTestCase(AbstractTestCase):
    @property
    def stored_event_repo(self):
        """
        Returns a stored event repository.

        Concrete log test cases will provide this method.
        """
        raise NotImplementedError

    def setUp(self):
        super(LogTestCase, self).setUp()

        # Check we're starting clean, event handler-wise.
        assert_event_handlers_empty()

        # Setup the persistence subscriber.
        self.event_store = EventStore(self.stored_event_repo)
        self.persistence_subscriber = PersistenceSubscriber(event_store=self.event_store)

        self.log_repo = LogRepo(self.event_store)

    def tearDown(self):
        # Close the persistence subscriber.
        self.persistence_subscriber.close()

        super(LogTestCase, self).tearDown()

        # Check we finished clean, event handler-wise.
        assert_event_handlers_empty()

    def test_entity_lifecycle(self):
        log = self.log_repo.get_or_create(log_name='log1', bucket_size='year')
        self.assertIsInstance(log, Log)
        self.assertEqual(log.name, 'log1')
        self.assertEqual(log.bucket_size, 'year')

        # Test get_logger and get_log_reader().
        logger = get_logger(log)
        self.assertIsInstance(logger, Logger)
        message1 = 'This is message 1'
        message2 = 'This is message 2'
        message3 = 'This is message 3'
        message4 = 'This is message 4'
        message5 = 'This is message 5'
        message6 = 'This is message 6'
        event1 = logger.info(message1)
        event2 = logger.info(message2)
        event3 = logger.info(message3)
        halfway = uuid1().hex
        event4 = logger.info(message4)
        event5 = logger.info(message5)
        event6 = logger.info(message6)

        # Check we can get all the messages (query running in descending order).
        log_reader = get_log_reader(log, event_store=self.event_store)
        messages = list(log_reader.get_messages(is_ascending=False))
        self.assertEqual(len(messages), 6)
        self.assertEqual(messages[0], message6)
        self.assertEqual(messages[1], message5)
        self.assertEqual(messages[2], message4)
        self.assertEqual(messages[3], message3)
        self.assertEqual(messages[4], message2)
        self.assertEqual(messages[5], message1)

        # Check we can get all the messages (query running in ascending order).
        messages = list(log_reader.get_messages(is_ascending=True))
        self.assertEqual(len(messages), 6)
        self.assertEqual(messages[0], message1)
        self.assertEqual(messages[1], message2)
        self.assertEqual(messages[2], message3)
        self.assertEqual(messages[3], message4)
        self.assertEqual(messages[4], message5)
        self.assertEqual(messages[5], message6)

        # Check we can get messages after halfway (query running in descending order).
        messages = list(log_reader.get_messages(after=halfway, is_ascending=False))
        self.assertEqual(len(messages), 3)
        self.assertEqual(messages[0], message6)
        self.assertEqual(messages[1], message5)
        self.assertEqual(messages[2], message4)

        # Check we can get messages until halfway (query running in descending order).
        messages = list(log_reader.get_messages(until=halfway, is_ascending=False))
        self.assertEqual(len(messages), 3)
        self.assertEqual(messages[0], message3)
        self.assertEqual(messages[1], message2)
        self.assertEqual(messages[2], message1)

        # Check we can get messages until halfway (query running in ascending order).
        messages = list(log_reader.get_messages(until=halfway, is_ascending=True))
        self.assertEqual(len(messages), 3)
        self.assertEqual(messages[0], message1)
        self.assertEqual(messages[1], message2)
        self.assertEqual(messages[2], message3)

        # Check we can get messages after halfway (query running in ascending order).
        messages = list(log_reader.get_messages(after=halfway, is_ascending=True))
        self.assertEqual(len(messages), 3)
        self.assertEqual(messages[0], message4)
        self.assertEqual(messages[1], message5)
        self.assertEqual(messages[2], message6)

        # Check we can get last three messages (query running in descending order).
        messages = list(log_reader.get_messages(limit=3, is_ascending=False))
        self.assertEqual(len(messages), 3)
        self.assertEqual(messages[0], message6)
        self.assertEqual(messages[1], message5)
        self.assertEqual(messages[2], message4)

        # Check we can get first three messages (query running in ascending order).
        messages = list(log_reader.get_messages(limit=3, is_ascending=True))
        self.assertEqual(len(messages), 3)
        self.assertEqual(messages[0], message1)
        self.assertEqual(messages[1], message2)
        self.assertEqual(messages[2], message3)

        # Check we can get last line (query running in descending order).
        messages = list(log_reader.get_messages(limit=1, after=halfway, is_ascending=False))
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0], message6)

        # Check we can get the first line after halfway (query running in ascending order).
        messages = list(log_reader.get_messages(limit=1, after=halfway, is_ascending=True))
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0], message4)

        # Check we can get the first line before halfway (query running in descending order).
        messages = list(log_reader.get_messages(limit=1, until=halfway, is_ascending=False))
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0], message3)

        # Check we can get the first line (query running in ascending order).
        messages = list(log_reader.get_messages(limit=1, until=halfway, is_ascending=True))
        self.assertEqual(len(messages), 1)
        self.assertEqual(messages[0], message1)

        # Check there isn't a line after the last line (query running in ascending order).
        messages = list(log_reader.get_messages(limit=1, after=event6.domain_event_id, is_ascending=True))
        self.assertEqual(len(messages), 0)

        # Check there is nothing somehow both after and until halfway.
        messages = list(log_reader.get_messages(after=halfway, until=halfway))
        self.assertEqual(len(messages), 0)

    def test_buckets(self):
        # Start new log.
        log = start_new_log(name='log1', bucket_size='second')

        # Write messages across the time interval
        start = datetime.datetime.now()
        logger = get_logger(log)
        number_of_messages = 300
        events = []
        for i in range(number_of_messages):
            message_logged = logger.info(str(i))
            events.append(message_logged)
            sleep(0.01)
        self.assertGreater(datetime.datetime.now() - start, datetime.timedelta(seconds=1))

        # Get the messages in descending order.
        reader = get_log_reader(log, self.event_store)
        messages = list(reader.get_messages(is_ascending=False, page_size=10))
        self.assertEqual(len(messages), number_of_messages)

        # Expect the order of the messages is the reverse of the created order.
        self.assertEqual(messages, list(reversed([str(i) for i in range(number_of_messages)])))

        # Get the messages in ascending order.
        messages = list(reader.get_messages(is_ascending=True, page_size=10))
        self.assertEqual(len(messages), number_of_messages)

        # Expect the order of the messages is the same as the created order.
        self.assertEqual(messages, [str(i) for i in range(number_of_messages)])

        # Get a limited number of messages in descending order.
        limit = 150
        messages = list(reader.get_messages(is_ascending=False, page_size=10, limit=limit))
        self.assertLess(limit, number_of_messages)
        self.assertEqual(len(messages), limit)

        # Expect the order of the messages is the reverse of the created order.
        self.assertEqual(messages, list(reversed([str(i) for i in range(number_of_messages)]))[:limit])

        # Get a limited number of messages in ascending order.
        limit = 150
        messages = list(reader.get_messages(is_ascending=True, page_size=10, limit=limit))
        self.assertLess(limit, number_of_messages)
        self.assertEqual(len(messages), limit)

        # Expect the order of the messages is the same as the created order.
        self.assertEqual(messages, [str(i) for i in range(limit)])

        # Get a limited number of messages in descending order from the midpoint down.
        limit = 110
        midpoint = events[150].domain_event_id
        messages = list(reader.get_messages(is_ascending=False, page_size=10, limit=limit, until=midpoint))
        self.assertLess(limit, number_of_messages)
        self.assertEqual(len(messages), limit)

        # Expect the order of the messages is the reverse of the created order.
        self.assertEqual(messages, list(reversed([str(i) for i in range(150 - limit, 150)])))

        # Get a limited number of messages in ascending order from the midpoint up.
        limit = 110
        midpoint = events[149].domain_event_id
        messages = list(reader.get_messages(is_ascending=True, page_size=10, limit=limit, after=midpoint))
        self.assertLess(limit, number_of_messages)
        self.assertEqual(len(messages), limit)

        # Expect the order of the messages is the same as the created order.
        self.assertEqual(messages, [str(i) for i in range(150, 150 + limit)])

        # Get a limited number of messages in descending order above the midpoint down.
        limit = 200
        midpoint = events[150].domain_event_id
        messages = list(reader.get_messages(is_ascending=False, page_size=10, limit=limit, after=midpoint))
        self.assertLess(limit, number_of_messages)
        self.assertEqual(len(messages), 150)

        # Expect the order of the messages is the reverse of the created order.
        self.assertEqual(messages, list(reversed([str(i) for i in range(150, 300)])))

        # Get a limited number of messages in ascending order below the midpoint up.
        limit = 200
        midpoint = events[149].domain_event_id
        messages = list(reader.get_messages(is_ascending=True, page_size=10, limit=limit, until=midpoint))
        self.assertLess(limit, number_of_messages)
        self.assertEqual(len(messages), 150)

        # Expect the order of the messages is the same as the created order.
        self.assertEqual(messages, [str(i) for i in range(150)])

        #
        # Use the last position to start part way through.
        limit = 20
        last_position = reader.position
        messages = reader.get_messages(is_ascending=True, page_size=10, limit=limit, after=last_position)
        messages = list(messages)
        self.assertEqual(len(messages), limit)

        # Expect the order of the messages is the same as the created order.
        self.assertEqual(messages, [str(i) for i in range(150, 150 + limit)])

        # Do it again.
        last_position = reader.position
        messages = reader.get_messages(is_ascending=True, page_size=10, limit=limit, after=last_position)
        messages = list(messages)
        self.assertEqual(len(messages), limit)

        # Expect the order of the messages is the same as the created order.
        self.assertEqual(messages, [str(i) for i in range(150 + limit, 150 + limit * 2)])

        # Go back.
        last_position = reader.position
        messages = reader.get_messages(is_ascending=False, page_size=10, limit=limit, until=last_position)
        messages = list(messages)
        self.assertEqual(len(messages), limit)

        # Expect the order of the messages is the reverse of the created order.
        self.assertEqual(messages, [str(i) for i in range(148 + limit * 2, 148 + limit, -1)])

        # Go back.
        last_position = reader.position
        messages = reader.get_messages(is_ascending=False, page_size=10, limit=limit, until=last_position)
        messages = list(messages)
        self.assertEqual(len(messages), limit)

        # Expect the order of the messages is the reverse of the created order.
        self.assertEqual(messages, [str(i) for i in range(128 + limit * 2, 128 + limit, -1)])

        # Go back.
        last_position = reader.position
        messages = reader.get_messages(is_ascending=False, page_size=10, limit=limit, until=last_position)
        messages = list(messages)
        self.assertEqual(len(messages), limit)

        # Expect the order of the messages is the reverse of the created order.
        self.assertEqual(messages, [str(i) for i in range(108 + limit * 2, 108 + limit, -1)])

        # Repeat.
        messages = reader.get_messages(is_ascending=False, page_size=10, limit=limit, until=last_position)
        messages = list(messages)
        self.assertEqual(len(messages), limit)

        # Expect the order of the messages is the reverse of the created order.
        self.assertEqual(messages, [str(i) for i in range(108 + limit * 2, 108 + limit, -1)])


class TestLogWithCassandra(CassandraStoredEventRepoTestCase, LogTestCase):
    pass


class TestLogWithPythonObjects(PythonObjectsTestCase, LogTestCase):
    pass


class TestLogWithSQLAlchemy(SQLAlchemyTestCase, LogTestCase):
    pass



from eventsourcing.application.example.base import ExampleApplication
from eventsourcing.application.with_sqlalchemy import EventSourcingWithSQLAlchemy


class ExampleApplicationWithSQLAlchemy(EventSourcingWithSQLAlchemy, ExampleApplication):
    """
    Example event sourced application.

    This application has an Example repository, and a factory method for
    registering new examples. It inherits an event store, a persistence
    subscriber, and a stored event repository, and a database connection.
    """
    pass



from eventsourcing.application.base import EventSourcingApplication
from eventsourcing.infrastructure.stored_event_repos.with_cassandra import setup_cassandra_connection, \
    get_cassandra_setup_params, CassandraStoredEventRepository, DEFAULT_CASSANDRA_KEYSPACE, DEFAULT_CASSANDRA_HOSTS, \
    DEFAULT_CASSANDRA_CONSISTENCY_LEVEL, DEFAULT_CASSANDRA_PORT, DEFAULT_CASSANDRA_PROTOCOL_VERSION


class EventSourcingWithCassandra(EventSourcingApplication):

    def __init__(self, hosts=DEFAULT_CASSANDRA_HOSTS, consistency=DEFAULT_CASSANDRA_CONSISTENCY_LEVEL,
                 default_keyspace=DEFAULT_CASSANDRA_KEYSPACE, port=DEFAULT_CASSANDRA_PORT,
                 protocol_version=DEFAULT_CASSANDRA_PROTOCOL_VERSION, username=None, password=None, **kwargs):
        self.setup_cassandra_connection(hosts, consistency, default_keyspace, port, protocol_version, username,
                                        password)
        super(EventSourcingWithCassandra, self).__init__(**kwargs)

    @staticmethod
    def setup_cassandra_connection(*args):
        setup_cassandra_connection(*get_cassandra_setup_params(*args))

    def create_stored_event_repo(self, **kwargs):
        return CassandraStoredEventRepository(**kwargs)

    def close(self):
        super(EventSourcingWithCassandra, self).close()
        # shutdown_cassandra_connection()



from eventsourcing.contrib.suffixtrees.domain.model.suffixtree import SuffixTree, SuffixTreeRepository, NodeRepository, Node, Edge, \
    EdgeRepository
from eventsourcing.infrastructure.event_sourced_repo import EventSourcedRepository


class SuffixTreeRepo(EventSourcedRepository, SuffixTreeRepository):
    domain_class = SuffixTree


class NodeRepo(EventSourcedRepository, NodeRepository):
    domain_class = Node


class EdgeRepo(EventSourcedRepository, EdgeRepository):
    domain_class = Edge



# from eventsourcing.infrastructure.stored_events.base import StoredEventRepository
from eventsourcing.infrastructure.stored_event_repos.with_python_objects import PythonObjectsStoredEventRepository
# from eventsourcing.infrastructure.stored_event_repos.transcoders import StoredEvent


# Todo: Figure out if this is actually needed by anything anywhere (no currently visible usages).
class SharedMemoryStoredEventRepository(PythonObjectsStoredEventRepository):
    #
    # def __init__(self):
    #     super(SharedMemoryStoredEventRepository, self).__init__()
    #     self._by_id = {}
    #     self._by_stored_entity_id = {}
    #     self._by_topic = {}
    #
    # def append(self, stored_event):
    #     assert isinstance(stored_event, StoredEvent)
    #     event_id = stored_event.event_id
    #     stored_entity_id = stored_event.stored_entity_id
    #     topic = stored_event.event_topic
    #
    #     # Index by event ID.
    #     self._by_id[event_id] = stored_event
    #
    #     # Index by entity ID.
    #     if stored_entity_id not in self._by_stored_entity_id:
    #         self._by_stored_entity_id[stored_entity_id] = []
    #     self._by_stored_entity_id[stored_entity_id].append(stored_event)
    #
    #     # Index by event topic.
    #     if topic not in self._by_topic:
    #         self._by_topic[topic] = []
    #     self._by_topic[topic].append(stored_event)
    #
    # def __getitem__(self, event_id):
    #     return self._by_id[event_id]
    #
    # def __contains__(self, event_id):
    #     return event_id in self._by_id
    #
    # def get_entity_events(self, stored_entity_id, after=None, until=None, limit=None):
    #     if stored_entity_id not in self._by_stored_entity_id:
    #         return []
    #     else:
    #         return self._by_stored_entity_id[stored_entity_id]
    #
    # def get_most_recent_event(self, stored_entity_id):
    #     if stored_entity_id not in self._by_stored_entity_id:
    #         return None
    #     else:
    #         return self._by_stored_entity_id[stored_entity_id][-1]
    #
    # def get_topic_events(self, event_topic):
    #     return reversed(self._by_topic[event_topic])
    pass



import importlib
import itertools
from abc import ABCMeta
from collections import OrderedDict
from uuid import uuid1

from six import with_metaclass

from eventsourcing.exceptions import TopicResolutionError
from eventsourcing.utils.time import timestamp_from_uuid


class QualnameABCMeta(ABCMeta):
    """Supplies __qualname__ to object classes with this metaclass.
    """
    __outer_classes = {}

    if not hasattr(object, '__qualname__'):

        def __init__(cls, name, bases, dict):
            super(QualnameABCMeta, cls).__init__(name, bases, dict)
            QualnameABCMeta.__outer_classes[cls] = None
            for class_attr in dict.values():  # iterate class attributes
                for outer_class in QualnameABCMeta.__outer_classes:
                    if class_attr == outer_class:  # is the object an already registered type?
                        QualnameABCMeta.__outer_classes[outer_class] = cls
                        break

        @property
        def __qualname__(cls):
            c = QualnameABCMeta.__outer_classes[cls]
            if c is None:
                return cls.__name__
            else:
                return "%s.%s" % (c.__qualname__, cls.__name__)


def create_domain_event_id():
    return uuid1().hex


class DomainEvent(with_metaclass(QualnameABCMeta)):

    always_encrypt = False

    def __init__(self, entity_id, entity_version, domain_event_id=None, **kwargs):
        self.__dict__['entity_id'] = entity_id
        self.__dict__['entity_version'] = entity_version
        self.__dict__['domain_event_id'] = domain_event_id or create_domain_event_id()
        self.__dict__.update(kwargs)

    def __setattr__(self, key, value):
        raise AttributeError("DomainEvent attributes are read-only")

    @property
    def entity_id(self):
        return self.__dict__['entity_id']

    @property
    def entity_version(self):
        return self.__dict__['entity_version']

    @property
    def domain_event_id(self):
        return self.__dict__['domain_event_id']

    @property
    def timestamp(self):
        return timestamp_from_uuid(self.__dict__['domain_event_id'])

    def __eq__(self, rhs):
        if type(self) is not type(rhs):
            return NotImplemented
        return self.__dict__ == rhs.__dict__

    def __ne__(self, rhs):
        return not (self == rhs)

    def __hash__(self):
        return hash(tuple(itertools.chain(sorted(self.__dict__.items()), [type(self)])))

    def __repr__(self):
        return self.__class__.__qualname__ + "(" + ', '.join(
            "{0}={1!r}".format(*item) for item in sorted(self.__dict__.items())) + ')'


_event_handlers = OrderedDict()


def subscribe(event_predicate, subscriber):
    if event_predicate not in _event_handlers:
        _event_handlers[event_predicate] = []
    _event_handlers[event_predicate].append(subscriber)


def unsubscribe(event_predicate, subscriber):
    if event_predicate in _event_handlers:
        handlers = _event_handlers[event_predicate]
        if subscriber in handlers:
            handlers.remove(subscriber)
            if not handlers:
                _event_handlers.pop(event_predicate)


def publish(event):
    matching_handlers = []
    for event_predicate, handlers in _event_handlers.items():
        if event_predicate(event):
            for handler in handlers:
                if handler not in matching_handlers:
                    matching_handlers.append(handler)
    for handler in matching_handlers:
        handler(event)


class EventHandlersNotEmptyError(Exception):
    pass


def assert_event_handlers_empty():
    if len(_event_handlers):
        msg = "Event handlers are still subscribed: %s" % _event_handlers
        raise EventHandlersNotEmptyError(msg)


def topic_from_domain_class(domain_class):
    """Returns a string describing a domain event class.

    Args:
        domain_event: A domain event object.

    Returns:
        A string describing the domain event object's class.
    """
    # assert isinstance(domain_event, DomainEvent)
    return domain_class.__module__ + '#' + domain_class.__qualname__


def resolve_domain_topic(topic):
    """Return domain class described by given topic.

    Args:
        topic: A string describing a domain class.

    Returns:
        A domain class.

    Raises:
        TopicResolutionError: If there is no such domain class.
    """
    try:
        module_name, _, class_name = topic.partition('#')
        module = importlib.import_module(module_name)
    except ImportError as e:
        raise TopicResolutionError("{}: {}".format(topic, e))
    try:
        cls = resolve_attr(module, class_name)
    except AttributeError as e:
        raise TopicResolutionError("{}: {}".format(topic, e))
    return cls


def resolve_attr(obj, path):
    """A recursive version of getattr for navigating dotted paths.

    Args:
        obj: An object for which we want to retrieve a nested attribute.
        path: A dot separated string containing zero or more attribute names.

    Returns:
        The attribute referred to by obj.a1.a2.a3...

    Raises:
        AttributeError: If there is no such attribute.
    """
    if not path:
        return obj
    head, _, tail = path.partition('.')
    head_obj = getattr(obj, head)
    return resolve_attr(head_obj, tail)



from functools import reduce

from eventsourcing.domain.model.entity import EventSourcedEntity
from eventsourcing.domain.model.snapshot import AbstractSnapshop
from eventsourcing.domain.services.eventstore import AbstractEventStore
from eventsourcing.domain.services.snapshotting import AbstractSnapshotStrategy, entity_from_snapshot
from eventsourcing.domain.services.transcoding import make_stored_entity_id


class EventPlayer(object):
    """Reconstitutes domain entities from domain events
    retrieved from the event store, optionally with snapshots.
    """

    def __init__(self, event_store, id_prefix, mutate_func, page_size=None, is_short=False, snapshot_strategy=None):
        assert isinstance(event_store, AbstractEventStore), event_store
        if snapshot_strategy is not None:
            assert isinstance(snapshot_strategy, AbstractSnapshotStrategy)
        self.event_store = event_store
        self.id_prefix = id_prefix
        self.mutate_func = mutate_func
        self.page_size = page_size
        self.is_short = is_short
        self.snapshot_strategy = snapshot_strategy

    def replay_events(self, entity_id, after=None, until=None, initial_state=None, query_descending=False):
        """Reconstitutes a domain entity from stored events.
        """

        # Make the stored entity ID.
        stored_entity_id = self.make_stored_entity_id(entity_id)

        if self.is_short and after is None and until is None and self.page_size is None:
            # Speed up for events are stored in descending order (e.g. in Cassandra).
            # - the inclusiveness or exclusiveness of until and after,
            #   and the end of the stream that is limited depends on
            #   query_ascending, so we can't use this method with them
            # - also if there's a page size, it probably isn't short
            is_ascending=False
        else:
            is_ascending = not query_descending

        # Get entity's domain events from the event store.
        domain_events = self.event_store.get_entity_events(
            stored_entity_id=stored_entity_id,
            after=after,
            until=until,
            page_size=self.page_size,
            is_ascending=is_ascending,
        )

        if not is_ascending:
            domain_events = reversed(list(domain_events))

        # Copy initial state, to preserve state of given object.
        if initial_state is not None:
            initial_state_copy = object.__new__(type(initial_state))
            initial_state_copy.__dict__.update(initial_state.__dict__)
            initial_state = initial_state_copy

        # Mutate initial state using the sequence of domain events.
        domain_entity = reduce(self.mutate_func, domain_events, initial_state)

        # Return the resulting domain entity.
        return domain_entity

    def take_snapshot(self, entity_id, until=None):
        """Takes a snapshot of the entity as it existed after
        the most recent event, optionally until a given time.
        """
        if self.snapshot_strategy is None:
            raise ValueError("Can't take snapshots without a snapshot strategy.")

        # Get the last event (optionally until a particular time).
        last_event = self.get_most_recent_event(entity_id, until=until)

        # If there aren't any events, there can't be a snapshot, so return None.
        if last_event is None:
            return None

        # If there is something to snapshot, then look for
        # the last snapshot before the last event.
        last_event_id = last_event.domain_event_id
        last_snapshot = self.get_snapshot(entity_id, until=last_event_id)

        # If there's a snapshot...
        if last_snapshot:
            assert isinstance(last_snapshot, AbstractSnapshop), type(last_snapshot)

            # and it is coincident with the last event...
            if last_snapshot.at_event_id == last_event_id:
                # then there's nothing to do.
                return last_snapshot
            else:
                # Otherwise there must be events after the snapshot, so
                # remember to get events after the last event that was
                # applied to the entity, and obtain the initial entity
                # state so those event can be applied.
                after_event_id = last_snapshot.at_event_id
                initial_state = entity_from_snapshot(last_snapshot)
        else:
            # If there isn't a snapshot, start from scratch.
            after_event_id = None
            initial_state = None

        # Get entity in the state after this event was applied.
        entity = self.replay_events(entity_id, after=after_event_id, until=last_event_id, initial_state=initial_state)

        # Take a snapshot of the entity.
        return self.snapshot_strategy.take_snapshot(entity, at_event_id=last_event_id)

    def get_snapshot(self, entity_id, until=None):
        """Uses strategy to get a snapshot of the entity,
        optional as it existed until a given time.
        """
        if self.snapshot_strategy:
            stored_entity_id = self.make_stored_entity_id(entity_id=entity_id)
            return self.snapshot_strategy.get_snapshot(stored_entity_id, until=until)

    def get_most_recent_event(self, entity_id, until=None):
        """Returns the most recent event for the given entity ID.
        """
        stored_entity_id = self.make_stored_entity_id(entity_id)
        return self.event_store.get_most_recent_event(stored_entity_id, until=until)

    def make_stored_entity_id(self, entity_id):
        """Prefixes the given entity ID with the ID prefix for entity's type.
        """
        return make_stored_entity_id(self.id_prefix, entity_id)

    def fastforward(self, stale_entity, until=None):
        assert isinstance(stale_entity, EventSourcedEntity)

        # Get the entity version object, which provides the event ID.
        stored_entity_id = self.make_stored_entity_id(stale_entity.id)

        # The last applied event version is the one before the current
        # version number, because each mutator increments the version
        # after an event is applied, and it starts by checking the
        # version number is the same. Hence version of the last applied
        # is the value of the entity's version property minus one.
        last_applied_version_number = stale_entity.version - 1
        last_applied_entity_version = self.event_store.get_entity_version(
            stored_entity_id=stored_entity_id,
            version=last_applied_version_number
        )
        last_applied_event_id = last_applied_entity_version.event_id

        # Replay the events since the entity version.
        fresh_entity = self.replay_events(
            entity_id=stale_entity.id,
            after=last_applied_event_id,
            until=until,
            initial_state=stale_entity,
        )

        # Return the fresh instance.
        return fresh_entity



from eventsourcing.application.example.with_pythonobjects import ExampleApplicationWithPythonObjects
from eventsourcing.tests.unit_test_cases_example_application import ExampleApplicationTestCase


class TestApplicationWithPythonObjects(ExampleApplicationTestCase):

    def create_app(self):
        return ExampleApplicationWithPythonObjects()



import unittest

from eventsourcing.domain.model.example import Example
from eventsourcing.domain.services.eventstore import EventStore
from eventsourcing.infrastructure.stored_event_repos.with_python_objects import PythonObjectsStoredEventRepository


class TestEventStore(unittest.TestCase):

    def test_get_entity_events(self):
        repo = PythonObjectsStoredEventRepository()
        event_store = EventStore(stored_event_repo=repo)

        # Check there are zero stored events in the repo.
        entity_events = event_store.get_entity_events(stored_entity_id='Example::entity1')
        entity_events = list(entity_events)
        self.assertEqual(0, len(entity_events))

        # Store a domain event.
        event1 = Example.Created(entity_id='entity1', a=1, b=2)
        event_store.append(event1)

        # Check there is one event in the event store.
        entity_events = event_store.get_entity_events(stored_entity_id='Example::entity1')
        entity_events = list(entity_events)
        self.assertEqual(1, len(entity_events))

        # Store another domain event.
        event1 = Example.AttributeChanged(entity_id='entity1', a=1, b=2, entity_version=1)
        event_store.append(event1)

        # Check there are two events in the event store.
        entity_events = event_store.get_entity_events(stored_entity_id='Example::entity1')
        entity_events = list(entity_events)
        self.assertEqual(2, len(entity_events))

        # Check there are two events in the event store.
        entity_events = event_store.get_entity_events(stored_entity_id='Example::entity1')
        entity_events = list(entity_events)
        self.assertEqual(2, len(entity_events))



from math import floor
from uuid import uuid1

import six

from eventsourcing.application.example.base import ExampleApplication
from eventsourcing.application.example.with_cassandra import ExampleApplicationWithCassandra
from eventsourcing.application.example.with_pythonobjects import ExampleApplicationWithPythonObjects
from eventsourcing.application.example.with_sqlalchemy import ExampleApplicationWithSQLAlchemy
from eventsourcing.domain.model.example import register_new_example, Example
from eventsourcing.domain.model.log import get_logger, start_new_log
from eventsourcing.domain.services.transcoding import make_stored_entity_id
from eventsourcing.infrastructure.log_reader import get_log_reader, LogReader
from eventsourcing.infrastructure.stored_event_repos.with_cassandra import create_cassandra_keyspace_and_tables, \
    drop_cassandra_keyspace
from eventsourcing.domain.services.cipher import AESCipher
from eventsourcing.tests.unit_test_cases import AbstractTestCase
from eventsourcing.tests.test_utils import utc_now


class PerformanceTestCase(AbstractTestCase):

    def setUp(self):
        super(PerformanceTestCase, self).setUp()
        self.app = self.create_app()
        assert isinstance(self.app, ExampleApplication)

    def create_app(self):
        """Returns instance of application object to be tested.

        Sub-classes are required to override this method to return an application object.
        """
        raise NotImplementedError

    def test_entity_performance(self):
        """
        Reports on the performance of Example entity and repo.

        NB: This test doesn't actually check anything, so it isn't really a test.
        """

        # Initialise dict of entities.
        self.entities = {}

        report_name = type(self).__name__[4:]
        print("\n\n{} report:\n".format(report_name))

        repetitions = 10

        # NB: Use range(1, 5) to test whether we can get more than 10000 event from Cassandra.
        for i in six.moves.range(0, 5):
            # Setup a number of entities, with different lengths of event history.
            payload = 3
            # payload = str([uuid4().hex for _ in six.moves.range(100000)])
            example = register_new_example(a=1, b=payload)
            self.entities[i] = example

            # Beat a number of times.
            start_beating = utc_now()
            num_beats = int(floor(10 ** i))
            for _ in six.moves.range(num_beats):
                example.beat_heart()
            time_beating = utc_now() - start_beating
            print("Time to beat {} times: {:.2f}s ({:.0f} beats/s, {:.6f}s each)"
                  "".format(num_beats, time_beating, num_beats / time_beating, time_beating / num_beats))

            # Get the last n events from the repo.
            def last_n(n):
                n = min(n, num_beats + 1)
                stored_entity_id = make_stored_entity_id('Example', example.id)
                repo = self.app.example_repo.event_player.event_store.stored_event_repo

                start_last_n = utc_now()
                last_n_stored_events = []
                for _ in six.moves.range(repetitions):
                    last_n_stored_events = repo.get_most_recent_events(stored_entity_id, limit=n)
                time_last_n = (utc_now() - start_last_n) / repetitions

                num_retrieved_events = len(list(last_n_stored_events))
                events_per_second = num_retrieved_events / time_last_n
                print(("Time to get last {:>"+str(i+1)+"} events after {} events: {:.6f}s ({:.0f} events/s)"
                      "").format(n, num_beats + 1, time_last_n, events_per_second))

            for j in range(0, i+1):
                last_n(10**j)

            # Get the entity by replaying all events (which it must since there isn't a snapshot).
            start_replay = utc_now()
            for _ in six.moves.range(repetitions):
                example = self.app.example_repo[example.id]
                assert isinstance(example, Example)
                heartbeats = example.count_heartbeats()
                assert heartbeats == num_beats, (heartbeats, num_beats)

            time_replaying = (utc_now() - start_replay) / repetitions
            print("Time to replay {} beats: {:.2f}s ({:.0f} beats/s, {:.6f}s each)"
                  "".format(num_beats, time_replaying, num_beats / time_replaying, time_replaying / num_beats))

            # Take snapshot, and beat heart a few more times.
            self.app.example_repo.event_player.take_snapshot(example.id, until=uuid1().hex)

            extra_beats = 4
            for _ in six.moves.range(extra_beats):
                example.beat_heart()
            num_beats += extra_beats

            # Get the entity using snapshot and replaying events since the snapshot.
            start_replay = utc_now()
            for _ in six.moves.range(repetitions):
                example = self.app.example_repo[example.id]
            time_replaying = (utc_now() - start_replay) / repetitions

            events_per_second = (extra_beats + 1) / time_replaying  # +1 for the snapshot event
            beats_per_second = num_beats / time_replaying
            print("Time to replay snapshot with {} extra beats: {:.6f}s ({:.0f} events/s, {:.0f} beats/s)"
                  "".format(extra_beats, time_replaying, events_per_second, beats_per_second))
            
            print("")

    def test_log_performance(self):
        log = start_new_log('example', bucket_size='year')
        logger = get_logger(log)
        log_reader = get_log_reader(log, self.app.event_store)

        # Write a load of messages.
        start_write = utc_now()
        number_of_messages = 111
        events = []
        for i in range(number_of_messages):
            event = logger.info('Logger message number {}'.format(i))
            events.append(event)
        time_to_write = (utc_now() - start_write)
        print("Time to log {} messages: {:.2f}s ({:.0f} messages/s, {:.6f}s each)"
              "".format(number_of_messages, time_to_write, number_of_messages/ time_to_write,
                        time_to_write / number_of_messages))

        # Read pages of messages in descending order.
        # - get a limited number until a time, then use the earliest in that list as the position
        position = events[-1].domain_event_id

        page_size = 10

        # Page back through the log in reverse chronological order.
        previous_position = None
        next_position = None
        count_pages = 0
        total_time_to_read = 0
        total_num_reads = 0
        while True:
            start_read = utc_now()
            page_of_events, next_position = self.get_message_logged_events_and_next_position(log_reader, position, page_size)
            time_to_read = (utc_now() - start_read)
            total_time_to_read += time_to_read
            total_num_reads += 1
            count_pages += 1
            if next_position is None:
                break
            else:
                previous_position, position = position, next_position

        # Check we got to the end of the line.
        self.assertEqual(count_pages, 11)
        self.assertIsNone(next_position)
        self.assertTrue(previous_position)

        # Page forward through the log in chronological order.
        count_pages = 0
        position = None
        while True:
            start_read = utc_now()
            page_of_events, next_position = self.get_message_logged_events_and_next_position(log_reader, position, page_size, is_ascending=True)
            time_to_read = (utc_now() - start_read)
            total_time_to_read += time_to_read
            total_num_reads += 1
            count_pages += 1
            if next_position is None:
                break
            else:
                position = next_position

        self.assertEqual(count_pages, 11)
        self.assertIsNone(next_position)
        self.assertTrue(previous_position)

        reads_per_second = total_num_reads / total_time_to_read
        messages_per_second = reads_per_second * number_of_messages
        print("Time to read {} pages of logged messages: {:.6f}s ({:.0f} pages/s, {:.0f} messages/s))"
              "".format(total_num_reads, total_time_to_read, reads_per_second, messages_per_second))

    def get_message_logged_events_and_next_position(self, log_reader, position, page_size, is_ascending=False):
        assert isinstance(log_reader, LogReader), type(log_reader)
        assert isinstance(position, (six.string_types, type(None))), type(position)
        assert isinstance(page_size, six.integer_types), type(page_size)
        assert isinstance(is_ascending, bool)
        if is_ascending:
            after = position
            until = None
        else:
            after = None
            until = position

        events = log_reader.get_events(after=after, until=until, limit=page_size + 1, is_ascending=is_ascending)
        events = list(events)
        if len(events) == page_size + 1:
            next_position = events.pop().domain_event_id
        else:
            next_position = None
        return events, next_position


class TestCassandraPerformance(PerformanceTestCase):

    def create_app(self):
        return ExampleApplicationWithCassandra()

    def setUp(self):
        super(TestCassandraPerformance, self).setUp()

        # Setup the keyspace and column family for stored events.
        create_cassandra_keyspace_and_tables()

    def tearDown(self):
        # Drop the keyspace.
        drop_cassandra_keyspace()

        # Close the application.
        self.app.close()

        super(TestCassandraPerformance, self).tearDown()


class TestEncryptionPerformance(TestCassandraPerformance):

    def create_app(self):
        cipher = AESCipher(aes_key='0123456789abcdef')
        return ExampleApplicationWithCassandra(cipher=cipher, always_encrypt_stored_events=True)


class TestSQLAlchemyPerformance(PerformanceTestCase):

    def create_app(self):
        return ExampleApplicationWithSQLAlchemy(db_uri='sqlite:///:memory:')

    def tearDown(self):
        # Close the application.
        self.app.close()


class TestPythonObjectsPerformance(PerformanceTestCase):

    def create_app(self):
        # Setup the example application.
        return ExampleApplicationWithPythonObjects()

    def tearDown(self):
        # Close the application.
        self.app.close()



from eventsourcing.tests.unit_test_cases import BasicStoredEventRepositoryTestCase, \
    ConcurrentStoredEventRepositoryTestCase, SimpleStoredEventIteratorTestCase, ThreadedStoredEventIteratorTestCase
from eventsourcing.tests.unit_test_cases_sqlalchemy import SQLAlchemyTestCase


class TestSQLAlchemyStoredEventRepository(SQLAlchemyTestCase, BasicStoredEventRepositoryTestCase):
    pass


class TestSimpleStoredEventIteratorWithSQLAlchemy(SQLAlchemyTestCase, SimpleStoredEventIteratorTestCase):
    pass


class TestThreadedStoredEventIteratorWithSQLAlchemy(SQLAlchemyTestCase, ThreadedStoredEventIteratorTestCase):
    pass


class TestConcurrentStoredEventRepositoryWithSQLAlchemy(SQLAlchemyTestCase, ConcurrentStoredEventRepositoryTestCase):
    pass



from eventsourcing.domain.model.log import Log, LogRepository, start_new_log
from eventsourcing.exceptions import RepositoryKeyError
from eventsourcing.infrastructure.event_sourced_repo import EventSourcedRepository


class LogRepo(EventSourcedRepository, LogRepository):
    """
    Event sourced repository for the Example domain model entity.
    """
    domain_class = Log

    def get_or_create(self, log_name, bucket_size):
        """
        Gets or creates a log.

        :rtype: Log
        """
        try:
            return self[log_name]
        except RepositoryKeyError:
            return start_new_log(log_name, bucket_size=bucket_size)



import unittest

from eventsourcing.infrastructure.stored_event_repos.with_python_objects import PythonObjectsStoredEventRepository


class PythonObjectsTestCase(unittest.TestCase):

    @property
    def stored_event_repo(self):
        try:
            return self._stored_event_repo
        except AttributeError:
            stored_event_repo = PythonObjectsStoredEventRepository()
            self._stored_event_repo = stored_event_repo
            return stored_event_repo


class EventSourcingError(Exception):
    """
    Base class for explicit errors in the eventsourcing mechanism.
    """
    pass


class TopicResolutionError(EventSourcingError):
    """
    Error resolving a topic to a Python class.
    """
    pass


class EntityVersionDoesNotExist(EventSourcingError):
    """
    Error when accessing an entity version.
    """
    pass


class ConcurrencyError(EventSourcingError):
    """
    Error appending events to a versioned stream.
    """
    pass


class ConsistencyError(EventSourcingError):
    """
    Error applying an event stream to a versioned entity.
    """
    pass


class ProgrammingError(EventSourcingError):
    """
    Error when reaching conditions that must have arisen from programming errors.
    """
    pass

class RepositoryKeyError(KeyError, EventSourcingError):
    """
    Error when getting an entity from a repo.
    """
    pass



# coding=utf-8
from threading import Thread

from eventsourcing.domain.services.eventstore import AbstractStoredEventRepository, StoredEventIterator


class ThreadedStoredEventIterator(StoredEventIterator):
    def __iter__(self):
        # Start a thread to get a page of events.
        thread = self.start_thread()

        # Get pages of stored events, until the page isn't full.
        while True:
            # Wait for the next page of events.
            thread.join(timeout=30)

            # Count the page.
            self._inc_page_counter()

            # Get the stored events from the thread.
            stored_events = thread.stored_events

            # Count the number of stored events in this page.
            num_stored_events = len(stored_events)

            # Decide if this is the last page.
            is_last_page = num_stored_events != self.page_size

            if not is_last_page:
                # Update loop variables.
                position = stored_events[-1]
                self._update_position(position)

                # Start the next thread.
                thread = self.start_thread()

            # Yield each stored event.
            for stored_event in stored_events:

                # Stop if we're over the limit.
                if self.limit and self.all_event_counter >= self.limit:
                    raise StopIteration

                # Count each event.
                self._inc_all_event_counter()

                # Yield the event.
                yield stored_event

            # If that was the last page, then stop iterating.
            if is_last_page:
                raise StopIteration

    def start_thread(self):
        thread = GetEntityEventsThread(
            repo=self.repo,
            stored_entity_id=self.stored_entity_id,
            after=self.after,
            until=self.until,
            page_size=self.page_size,
            is_ascending=self.is_ascending
        )
        thread.start()
        return thread


class GetEntityEventsThread(Thread):
    def __init__(self, repo, stored_entity_id, after=None, until=None, page_size=None, is_ascending=True, *args,
                 **kwargs):
        super(GetEntityEventsThread, self).__init__(*args, **kwargs)
        assert isinstance(repo, AbstractStoredEventRepository)
        self.repo = repo
        self.stored_entity_id = stored_entity_id
        self.after = after
        self.until = until
        self.page_size = page_size
        self.is_ascending = is_ascending
        self.stored_events = None

    def run(self):
        self.stored_events = list(self.repo.get_entity_events(
            stored_entity_id=self.stored_entity_id,
            after=self.after,
            until=self.until,
            limit=self.page_size,
            query_ascending=self.is_ascending,
            results_ascending=self.is_ascending,
        ))



# coding=utf-8
from abc import ABCMeta, abstractmethod

import six

from eventsourcing.domain.model.events import DomainEvent
from eventsourcing.domain.services.transcoding import AbstractTranscoder, Transcoder, StoredEvent
from eventsourcing.exceptions import ProgrammingError, EntityVersionDoesNotExist, ConcurrencyError
from eventsourcing.utils.time import time_from_uuid


class AbstractEventStore(six.with_metaclass(ABCMeta)):

    @abstractmethod
    def append(self, domain_event):
        """
        Put domain event in event store for later retrieval.
        """

    @abstractmethod
    def get_entity_events(self, stored_entity_id, after=None, until=None, limit=None, is_ascending=True,
                          page_size=None):
        pass

    @abstractmethod
    def get_entity_version(self, stored_entity_id, version):
        pass

    @abstractmethod
    def get_most_recent_event(self, stored_entity_id, until=None):
        pass


class EventStore(AbstractEventStore):

    def __init__(self, stored_event_repo, transcoder=None):
        assert isinstance(stored_event_repo, AbstractStoredEventRepository), stored_event_repo
        if transcoder is None:
            transcoder = Transcoder()
        assert isinstance(transcoder, AbstractTranscoder), transcoder
        self.stored_event_repo = stored_event_repo

        self.transcoder = transcoder

    def append(self, domain_event):
        assert isinstance(domain_event, DomainEvent)
        # Serialize the domain event.
        stored_event = self.transcoder.serialize(domain_event)

        # Append the stored event to the stored event repo.
        self.stored_event_repo.append(
            new_stored_event=stored_event,
            new_version_number=domain_event.entity_version,
        )

    def get_entity_events(self, stored_entity_id, after=None, until=None, limit=None, is_ascending=True,
                          page_size=None):
        # Get the events that have been stored for the entity.
        if page_size:
            stored_events = self.stored_event_repo.iterate_entity_events(
                stored_entity_id=stored_entity_id,
                after=after,
                until=until,
                limit=limit,
                is_ascending=is_ascending,
                page_size=page_size
            )
        else:
            stored_events = self.stored_event_repo.get_entity_events(
                stored_entity_id=stored_entity_id,
                after=after,
                until=until,
                limit=limit,
                query_ascending=is_ascending,
                results_ascending=is_ascending,
            )

        # Deserialize all the stored event objects into domain event objects.
        return six.moves.map(self.transcoder.deserialize, stored_events)

    def get_most_recent_event(self, stored_entity_id, until=None):
        """Returns last event for given stored entity ID.

        :rtype: DomainEvent, NoneType
        """
        stored_event = self.stored_event_repo.get_most_recent_event(stored_entity_id, until=until)
        return None if stored_event is None else self.transcoder.deserialize(stored_event)

    def get_entity_version(self, stored_entity_id, version):
        return self.stored_event_repo.get_entity_version(stored_entity_id=stored_entity_id, version_number=version)


class AbstractStoredEventRepository(six.with_metaclass(ABCMeta)):

    def __init__(self, always_check_expected_version=False, always_write_entity_version=False):
        """
        Base class for a persistent collection of stored events.
        """
        self.always_check_expected_version = always_check_expected_version
        self.always_write_entity_version = always_write_entity_version
        if self.always_check_expected_version and not self.always_write_entity_version:
            raise ProgrammingError("If versions are checked, they must also be written.")

    def append(self, new_stored_event, new_version_number=None, max_retries=3, artificial_failure_rate=0):
        """
        Saves given stored event in this repository.
        """
        # Check the new event is a stored event instance.
        assert isinstance(new_stored_event, StoredEvent)

        # Validate the expected version.
        if self.always_check_expected_version:
            # noinspection PyTypeChecker
            self.validate_expected_version(new_stored_event, new_version_number)

        # Write the event.
        self.write_version_and_event(
            new_stored_event=new_stored_event,
            new_version_number=new_version_number,
            max_retries=max_retries,
            artificial_failure_rate=artificial_failure_rate,
        )

    def validate_expected_version(self, new_stored_event, new_version_number):
        """
        Checks the expected version exists and occurred before the new event.

        Raises a concurrency error if the expected version doesn't exist,
        or if the new event occurred before the expected version occurred.
        """
        assert isinstance(new_stored_event, StoredEvent)

        stored_entity_id = new_stored_event.stored_entity_id
        expected_version_number = self.decide_expected_version_number(new_version_number)
        if expected_version_number is not None:
            assert isinstance(expected_version_number, six.integer_types)
            try:
                entity_version = self.get_entity_version(
                    stored_entity_id=stored_entity_id,
                    version_number=expected_version_number
                )
            except EntityVersionDoesNotExist:
                raise ConcurrencyError("Expected version '{}' of stored entity '{}' not found."
                                       "".format(expected_version_number, stored_entity_id))
            else:
                if not time_from_uuid(new_stored_event.event_id) > time_from_uuid(entity_version.event_id):
                    raise ConcurrencyError("New event ID '{}' occurs before last version event ID '{}' for entity {}"
                                           "".format(new_stored_event.event_id, entity_version.event_id, stored_entity_id))

    def decide_expected_version_number(self, new_version_number):
        return new_version_number - 1 if new_version_number else None

    @abstractmethod
    def get_entity_version(self, stored_entity_id, version_number):
        """
        Returns entity version object for given entity version ID.

        :rtype: EntityVersion

        """

    @abstractmethod
    def write_version_and_event(self, new_stored_event, new_entity_version=None, max_retries=3, artificial_failure_rate=0):
        """
        Writes entity version and stored event into a database.
        """

    @abstractmethod
    def get_entity_events(self, stored_entity_id, after=None, until=None, limit=None, query_ascending=True,
                          results_ascending=True):
        """
        Returns all events for given entity ID in chronological order. Limit is max 10000.

        :rtype: list

        """

    def iterate_entity_events(self, stored_entity_id, after=None, until=None, limit=None, is_ascending=True,
                              page_size=None):
        """
        Returns all events for given entity ID by paging through the stored events.

        :rtype: list

        """
        return self.iterator_class(
            repo=self,
            stored_entity_id=stored_entity_id,
            page_size=min(page_size or limit, limit or page_size),
            after=after,
            until=until,
            limit=limit,
            is_ascending=is_ascending,
        )

    @property
    def iterator_class(self):
        return SimpleStoredEventIterator

    def get_most_recent_event(self, stored_entity_id, until=None):
        """
        Returns last event for given entity ID.

        :rtype: DomainEvent, NoneType

        """
        events = self.get_most_recent_events(stored_entity_id, until=until, limit=1)
        events = list(events)
        if len(events) == 1:
            return events[0]
        elif len(events) == 0:
            return None
        else:
            raise Exception("Shouldn't have more than one event object: {}"
                            "".format(events))

    def get_most_recent_events(self, stored_entity_id, until=None, limit=None):
        """
        Returns a stored event from a domain event.

        :rtype: list

        """
        return self.get_entity_events(
            stored_entity_id=stored_entity_id,
            until=until,
            limit=limit,
            query_ascending=False,
            results_ascending=False,
        )

    @staticmethod
    def map(func, iterable):
        return six.moves.map(func, iterable)

    @staticmethod
    def make_entity_version_id(stored_entity_id, version):
        """
        Constructs entity version ID from stored entity ID and version number.

        :rtype: str

        """
        return u"{}::version::{}".format(stored_entity_id, version)


class StoredEventIterator(six.with_metaclass(ABCMeta)):
    DEFAULT_PAGE_SIZE = 1000

    def __init__(self, repo, stored_entity_id, page_size=None, after=None, until=None, limit=None, is_ascending=True):
        assert isinstance(repo, AbstractStoredEventRepository), type(repo)
        assert isinstance(stored_entity_id, six.string_types)
        assert isinstance(page_size, (six.integer_types, type(None)))
        assert isinstance(limit, (six.integer_types, type(None)))
        self.repo = repo
        self.stored_entity_id = stored_entity_id
        self.page_size = page_size or self.DEFAULT_PAGE_SIZE
        self.after = after
        self.until = until
        self.limit = limit
        self.page_counter = 0
        self.all_event_counter = 0
        self.is_ascending = is_ascending

    def _inc_page_counter(self):
        """
        Increments the page counter.

        Each query result as a page, even if there are no items in the page. This really counts queries.
         - it is easy to divide the number of events by the page size if the "correct" answer is required
         - there will be a difference in the counts when the number of events can be exactly divided by the page
           size, because there is no way to know in advance that a full page is also the last page.
        """
        self.page_counter += 1

    def _inc_all_event_counter(self):
        self.all_event_counter += 1

    def _update_position(self, stored_event):
        if stored_event is None:
            return
        assert isinstance(stored_event, StoredEvent), type(stored_event)
        if self.is_ascending:
            self.after = stored_event.event_id
        else:
            self.until = stored_event.event_id

    @abstractmethod
    def __iter__(self):
        pass


class SimpleStoredEventIterator(StoredEventIterator):
    def __iter__(self):
        """
        Yields pages of events until the last page.

       """
        while True:
            # Get next page of events.
            if self.limit is not None:
                limit = min(self.page_size, self.limit - self.all_event_counter)
            else:
                limit = self.page_size

            if limit == 0:
                raise StopIteration

            # Get the events.
            stored_events = self.repo.get_entity_events(
                stored_entity_id=self.stored_entity_id,
                after=self.after,
                until=self.until,
                limit=limit,
                query_ascending=self.is_ascending,
                results_ascending=self.is_ascending,
            )

            # Count the page.
            self._inc_page_counter()

            # Start counting events in this page.
            in_page_event_counter = 0

            # Yield each stored event, so long as we aren't over the limit.
            for stored_event in stored_events:

                # Stop if we're over the limit.
                if self.limit and self.all_event_counter >= self.limit:
                    raise StopIteration

                # Count each event.
                self._inc_all_event_counter()
                in_page_event_counter += 1

                # Yield the event.
                yield stored_event

                # Remember the position as the last event.
                self._update_position(stored_event)

            # Decide if this is the last page.
            is_last_page = in_page_event_counter != self.page_size

            # If that was the last page, then stop iterating.
            if is_last_page:
                raise StopIteration


from unittest.case import TestCase

from eventsourcing.application.example.with_pythonobjects import ExampleApplicationWithPythonObjects


class App(ExampleApplicationWithPythonObjects):
    persist_events = False


class TestApplication(TestCase):

    def test_persistance_disabled(self):
        app = App()
        self.assertIsNone(app.persistence_subscriber)
        app.close()



from eventsourcing.application.base import EventSourcingApplication
from eventsourcing.domain.model.example import register_new_example
from eventsourcing.domain.services.snapshotting import EventSourcedSnapshotStrategy
from eventsourcing.infrastructure.event_sourced_repos.example_repo import ExampleRepo


class ExampleApplication(EventSourcingApplication):
    """
    Abstract example event sourced application.

    This application has an Example repository, and a factory method to construct new Example entities.

    It doesn't have a stored event repository.
    """
    def __init__(self, enable_occ=True, always_write_entity_version=True, **kwargs):
        super(ExampleApplication, self).__init__(
            enable_occ=enable_occ,
            always_write_entity_version=always_write_entity_version, **kwargs)
        self.snapshot_strategy = EventSourcedSnapshotStrategy(event_store=self.event_store)
        self.example_repo = ExampleRepo(event_store=self.event_store, snapshot_strategy=self.snapshot_strategy)

    def register_new_example(self, a, b):
        return register_new_example(a=a, b=b)



from unittest import TestCase

from eventsourcing.infrastructure.stored_event_repos.with_shared_memory import SharedMemoryStoredEventRepository
from eventsourcing.tests.unit_test_cases import BasicStoredEventRepositoryTestCase, SimpleStoredEventIteratorTestCase, \
    ThreadedStoredEventIteratorTestCase


class SharedMemoryTestCase(TestCase):

    @property
    def stored_event_repo(self):
        try:
            return self._stored_event_repo
        except AttributeError:
            stored_event_repo = SharedMemoryStoredEventRepository()
            self._stored_event_repo = stored_event_repo
            return self._stored_event_repo


class TestSharedMemoryStoredEventRepository(SharedMemoryTestCase, BasicStoredEventRepositoryTestCase):
    pass


class TestSimpleStoredEventIteratorWithPythonObjects(SharedMemoryTestCase, SimpleStoredEventIteratorTestCase):
    pass


class TestThreadedStoredEventIteratorWithPythonObjects(SharedMemoryTestCase, ThreadedStoredEventIteratorTestCase):
    pass



from abc import ABCMeta, abstractproperty

import six

from eventsourcing.domain.model.events import DomainEvent


class AbstractSnapshop(six.with_metaclass(ABCMeta)):

    @abstractproperty
    def topic(self):
        """Path to the class.
        """

    @abstractproperty
    def attrs(self):
        """Attributes of the instance.
        """

    @abstractproperty
    def at_event_id(self):
        """Last domain event ID that was applied to the snapshotted entity.
        """


class Snapshot(DomainEvent, AbstractSnapshop):

    def __init__(self, entity_id, topic, attrs, domain_event_id):
        super(Snapshot, self).__init__(entity_id=entity_id, topic=topic, attrs=attrs, domain_event_id=domain_event_id,
                                       entity_version=None)

    @property
    def topic(self):
        """Path to the class.
        """
        return self.__dict__['topic']

    @property
    def attrs(self):
        """Attributes of the instance.
        """
        return self.__dict__['attrs']

    @property
    def at_event_id(self):
        return self.__dict__['domain_event_id']



from time import sleep

import six
from sqlalchemy import create_engine
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.declarative.api import declarative_base
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.orm.scoping import ScopedSession
from sqlalchemy.sql.expression import asc, desc
from sqlalchemy.sql.schema import Column, Sequence
from sqlalchemy.sql.sqltypes import Integer, String, BigInteger, Text

from eventsourcing.domain.services.eventstore import AbstractStoredEventRepository
from eventsourcing.domain.services.transcoding import StoredEvent, EntityVersion
from eventsourcing.exceptions import ConcurrencyError, EntityVersionDoesNotExist
from eventsourcing.utils.time import timestamp_long_from_uuid


def get_scoped_session_facade(uri):
    if uri is None:
        raise ValueError
    engine = create_engine(uri, strategy='threadlocal')
    Base.metadata.create_all(engine)
    session_factory = sessionmaker(bind=engine)
    scoped_session_facade = scoped_session(session_factory)
    return scoped_session_facade


Base = declarative_base()


class SqlEntityVersion(Base):

    __tablename__ = 'entity_versions'

    entity_version_id = Column(String(355), primary_key=True)
    event_id = Column(String(255))


class SqlStoredEvent(Base):

    __tablename__ = 'stored_events'

    id = Column(Integer, Sequence('stored_event_id_seq'), primary_key=True)
    event_id = Column(String(255), index=True)
    timestamp_long = Column(BigInteger(), index=True)
    stored_entity_id = Column(String(255), index=True)
    event_topic = Column(String(255))
    event_attrs = Column(Text())


def from_sql(sql_stored_event):
    assert isinstance(sql_stored_event, SqlStoredEvent), sql_stored_event
    return StoredEvent(
        event_id=sql_stored_event.event_id,
        stored_entity_id=sql_stored_event.stored_entity_id,
        event_attrs=sql_stored_event.event_attrs,
        event_topic=sql_stored_event.event_topic
    )


def to_sql(stored_event):
    assert isinstance(stored_event, StoredEvent)
    return SqlStoredEvent(
        event_id=stored_event.event_id,
        timestamp_long=timestamp_long_from_uuid(stored_event.event_id),
        stored_entity_id=stored_event.stored_entity_id,
        event_attrs=stored_event.event_attrs,
        event_topic=stored_event.event_topic
    )


class SQLAlchemyStoredEventRepository(AbstractStoredEventRepository):

    def __init__(self, db_session, **kwargs):
        super(SQLAlchemyStoredEventRepository, self).__init__(**kwargs)
        assert isinstance(db_session, ScopedSession)
        self.db_session = db_session

    def write_version_and_event(self, new_stored_event, new_version_number=None, max_retries=3, artificial_failure_rate=0):
        """
        Writes new entity version and stored event into the database in a single transaction.
        """
        stored_entity_id = new_stored_event.stored_entity_id
        try:
            # Write entity version into the transaction.
            if self.always_write_entity_version and new_version_number is not None:
                assert isinstance(new_version_number, six.integer_types)
                new_entity_version_id = self.make_entity_version_id(stored_entity_id, new_version_number)
                new_entity_version = SqlEntityVersion(
                    entity_version_id=new_entity_version_id,
                    event_id=new_stored_event.event_id,
                )

                self.db_session.add(new_entity_version)

                if artificial_failure_rate:

                    # Todo: Add some retries and raise some artificial exceptions?
                    # if artificial_failure_rate and (random() > 1 - artificial_failure_rate):
                    #     raise Exception("Artificial failure")

                    # Commit the version number now to generate some contention.
                    #  - used to generate contention in tests
                    self.db_session.commit()
                    self.db_session.close()

                    # Increased latency here causes increased contention.
                    #  - used to generate contention in tests
                    sleep(artificial_failure_rate)

            # Write stored event into the transaction.
            self.db_session.add(to_sql(new_stored_event))

            # Commit the transaction.
            self.db_session.commit()

        except IntegrityError as e:
            # Rollback and raise concurrency error.
            self.db_session.rollback()
            raise ConcurrencyError("Version {} of entity {} already exists: {}"
                                   "".format(new_entity_version, stored_entity_id, e))
        except:
            # Rollback and reraise.
            self.db_session.rollback()
            raise
        finally:
            # Begin new transaction.
            self.db_session.close()

    def get_entity_version(self, stored_entity_id, version_number):
        entity_version_id = self.make_entity_version_id(stored_entity_id, version_number)
        version_number = self.db_session.query(SqlEntityVersion).filter_by(entity_version_id=entity_version_id).first()
        if version_number is None:
            raise EntityVersionDoesNotExist()
        assert isinstance(version_number, SqlEntityVersion)
        return EntityVersion(
            entity_version_id=entity_version_id,
            event_id=version_number.event_id,
        )

    def get_entity_events(self, stored_entity_id, after=None, until=None, limit=None, query_ascending=True,
                          results_ascending=True):
        try:
            query = self.db_session.query(SqlStoredEvent)
            query = query.filter_by(stored_entity_id=stored_entity_id)
            if query_ascending:
                query = query.order_by(asc(SqlStoredEvent.timestamp_long))
            else:
                query = query.order_by(desc(SqlStoredEvent.timestamp_long))

            if after is not None:
                if query_ascending:
                    query = query.filter(SqlStoredEvent.timestamp_long > timestamp_long_from_uuid(after))
                else:
                    query = query.filter(SqlStoredEvent.timestamp_long >= timestamp_long_from_uuid(after))

            if until is not None:
                if query_ascending:
                    query = query.filter(SqlStoredEvent.timestamp_long <= timestamp_long_from_uuid(until))
                else:
                    query = query.filter(SqlStoredEvent.timestamp_long < timestamp_long_from_uuid(until))

            if limit is not None:
                query = query.limit(limit)
            events = self.map(from_sql, query)
            events = list(events)
        finally:
            self.db_session.close()
        if results_ascending != query_ascending:
            events.reverse()
        return events



from eventsourcing.application.example.base import ExampleApplication
from eventsourcing.application.subscribers.persistence import PersistenceSubscriber
from eventsourcing.domain.model.example import Example
from eventsourcing.domain.services.eventstore import AbstractEventStore, AbstractStoredEventRepository
from eventsourcing.infrastructure.event_sourced_repos.example_repo import ExampleRepo
from eventsourcing.tests.unit_test_cases import AbstractTestCase


class ApplicationTestCase(AbstractTestCase):

    def create_app(self):
        raise NotImplementedError

    def setUp(self):
        super(ApplicationTestCase, self).setUp()
        self.app = self.create_app()

    def tearDown(self):
        self.app.close()
        super(ApplicationTestCase, self).tearDown()


class ExampleApplicationTestCase(ApplicationTestCase):

    def test(self):
        """
        Checks application works in the way an ExampleApplication should do.
        """

        # Check we're dealing with an example application.
        assert isinstance(self.app, ExampleApplication)

        # Check there's a stored event repo.
        self.assertIsInstance(self.app.stored_event_repo, AbstractStoredEventRepository)

        # Check there's an event store.
        self.assertIsInstance(self.app.event_store, AbstractEventStore)
        self.assertEqual(self.app.event_store.stored_event_repo, self.app.stored_event_repo)

        # Check there's a persistence subscriber.
        self.assertIsInstance(self.app.persistence_subscriber, PersistenceSubscriber)
        self.assertEqual(self.app.persistence_subscriber.event_store, self.app.event_store)

        # Check there's an example repository.
        self.assertIsInstance(self.app.example_repo, ExampleRepo)

        # Register a new example.
        example1 = self.app.register_new_example(a=10, b=20)
        self.assertIsInstance(example1, Example)

        # Check the example is available in the repo.
        entity1 = self.app.example_repo[example1.id]
        self.assertEqual(10, entity1.a)
        self.assertEqual(20, entity1.b)
        self.assertEqual(example1, entity1)

        # Change attribute values.
        entity1.a = 100

        # Check the new value is available in the repo.
        entity1 = self.app.example_repo[example1.id]
        self.assertEqual(100, entity1.a)



import time

from eventsourcing.exceptions import ConcurrencyError


def retry_after_concurrency_error(operation_to_retry, max_attempts=10, wait=0.02, args=()):
    tries = 0
    assert tries <= max_attempts
    while True:
        try:
            operation_to_retry(*args)
            break
        except ConcurrencyError as error:
            tries += 1
            if tries < max_attempts:
                time.sleep(wait * (1 + tries))
                continue
            else:
                raise error
